from . import processors
from .discord import (
    ApplicationCommandPermissionsUpdate,
    AutoModCreated,
    AutoModDeleted,
    AutoModExec,
    AutoModUpdated,
    BanCreate,
    BanRemove,
    BaseVoiceEvent,
    ChannelCreate,
    ChannelDelete,
    ChannelPinsUpdate,
    ChannelUpdate,
    GuildAuditLogEntryCreate,
    GuildAvailable,
    GuildEmojisUpdate,
    GuildJoin,
    GuildLeft,
    GuildMembersChunk,
    GuildScheduledEventCreate,
    GuildScheduledEventUpdate,
    GuildScheduledEventDelete,
    GuildScheduledEventUserAdd,
    GuildScheduledEventUserRemove,
    GuildStickersUpdate,
    GuildUnavailable,
    GuildUpdate,
    IntegrationCreate,
    IntegrationDelete,
    IntegrationUpdate,
    InteractionCreate,
    InviteCreate,
    InviteDelete,
    MemberAdd,
    MemberRemove,
    MemberUpdate,
    MessageCreate,
    MessageDelete,
    MessageDeleteBulk,
    MessageReactionAdd,
    MessageReactionRemove,
    MessageReactionRemoveAll,
    MessageUpdate,
    NewThreadCreate,
    PresenceUpdate,
    RoleCreate,
    RoleDelete,
    RoleUpdate,
    StageInstanceCreate,
    StageInstanceDelete,
    StageInstanceUpdate,
    ThreadCreate,
    ThreadDelete,
    ThreadListSync,
    ThreadMembersUpdate,
    ThreadMemberUpdate,
    ThreadUpdate,
    TypingStart,
    VoiceStateUpdate,
    VoiceUserDeafen,
    VoiceUserJoin,
    VoiceUserLeave,
    VoiceUserMove,
    VoiceUserMute,
    WebhooksUpdate,
)
from .internal import (
    AutocompleteCompletion,
    AutocompleteError,
    ButtonPressed,
    CallbackAdded,
    CommandCompletion,
    CommandError,
    Component,
    ComponentCompletion,
    ComponentError,
    Connect,
    Disconnect,
    Error,
    ExtensionCommandParse,
    ExtensionLoad,
    ExtensionUnload,
    Login,
    ModalCompletion,
    ModalError,
    Ready,
    Resume,
    Select,
    ShardConnect,
    ShardDisconnect,
    Startup,
    WebsocketReady,
)
from .base import BaseEvent, GuildEvent, RawGatewayEvent

__all__ = (
    "processors",
    "ApplicationCommandPermissionsUpdate",
    "AutocompleteCompletion",
    "AutocompleteError",
    "AutoModCreated",
    "AutoModDeleted",
    "AutoModExec",
    "AutoModUpdated",
    "BanCreate",
    "BanRemove",
    "BaseEvent",
    "BaseVoiceEvent",
    "ButtonPressed",
    "CallbackAdded",
    "ChannelCreate",
    "ChannelDelete",
    "ChannelPinsUpdate",
    "ChannelUpdate",
    "CommandCompletion",
    "CommandError",
    "Component",
    "ComponentCompletion",
    "ComponentError",
    "Connect",
    "Disconnect",
    "Error",
    "ExtensionCommandParse",
    "ExtensionLoad",
    "ExtensionUnload",
    "GuildAuditLogEntryCreate",
    "GuildAvailable",
    "GuildEmojisUpdate",
    "GuildEvent",
    "GuildJoin",
    "GuildLeft",
    "GuildMembersChunk",
    "GuildScheduledEventCreate",
    "GuildScheduledEventUpdate",
    "GuildScheduledEventDelete",
    "GuildScheduledEventUserAdd",
    "GuildScheduledEventUserRemove",
    "GuildStickersUpdate",
    "GuildUnavailable",
    "GuildUpdate",
    "IntegrationCreate",
    "IntegrationDelete",
    "IntegrationUpdate",
    "InteractionCreate",
    "InviteCreate",
    "InviteDelete",
    "Login",
    "MemberAdd",
    "MemberRemove",
    "MemberUpdate",
    "MessageCreate",
    "MessageDelete",
    "MessageDeleteBulk",
    "MessageReactionAdd",
    "MessageReactionRemove",
    "MessageReactionRemoveAll",
    "MessageUpdate",
    "ModalCompletion",
    "ModalError",
    "NewThreadCreate",
    "PresenceUpdate",
    "RawGatewayEvent",
    "Ready",
    "Resume",
    "RoleCreate",
    "RoleDelete",
    "RoleUpdate",
    "Select",
    "ShardConnect",
    "ShardDisconnect",
    "StageInstanceCreate",
    "StageInstanceDelete",
    "StageInstanceUpdate",
    "Startup",
    "ThreadCreate",
    "ThreadDelete",
    "ThreadListSync",
    "ThreadMembersUpdate",
    "ThreadMemberUpdate",
    "ThreadUpdate",
    "TypingStart",
    "VoiceStateUpdate",
    "VoiceUserDeafen",
    "VoiceUserJoin",
    "VoiceUserLeave",
    "VoiceUserMove",
    "VoiceUserMute",
    "WebhooksUpdate",
    "WebsocketReady",
)



from .activity import Activity, ActivityAssets, ActivityParty, ActivitySecrets, ActivityTimestamps
from .app_perms import ApplicationCommandPermission
from .application import Application
from .asset import Asset
from .auto_mod import AutoModerationAction, AutoModRule
from .channel import (
    BaseChannel,
    ChannelHistory,
    DM,
    DMChannel,
    DMGroup,
    GuildCategory,
    GuildChannel,
    GuildForum,
    GuildForumPost,
    GuildMedia,
    GuildNews,
    GuildNewsThread,
    GuildPrivateThread,
    GuildPublicThread,
    GuildStageVoice,
    GuildText,
    GuildVoice,
    InvitableMixin,
    MessageableMixin,
    PermissionOverwrite,
    process_permission_overwrites,
    ThreadableMixin,
    ThreadChannel,
    TYPE_ALL_CHANNEL,
    TYPE_CHANNEL_MAPPING,
    TYPE_DM_CHANNEL,
    TYPE_GUILD_CHANNEL,
    TYPE_MESSAGEABLE_CHANNEL,
    TYPE_THREAD_CHANNEL,
    TYPE_VOICE_CHANNEL,
    WebhookMixin,
)
from .color import (
    BrandColors,
    BrandColours,
    Color,
    COLOR_TYPES,
    Colour,
    FlatUIColors,
    FlatUIColours,
    MaterialColors,
    MaterialColours,
    process_color,
    process_colour,
    RoleColors,
    RoleColours,
)
from .components import (
    ActionRow,
    BaseComponent,
    BaseSelectMenu,
    Button,
    ChannelSelectMenu,
    get_components_ids,
    InteractiveComponent,
    MentionableSelectMenu,
    process_components,
    RoleSelectMenu,
    spread_to_rows,
    StringSelectMenu,
    StringSelectOption,
    TYPE_COMPONENT_MAPPING,
    UserSelectMenu,
)

from .embed import Embed, EmbedAttachment, EmbedAuthor, EmbedField, EmbedFooter, EmbedProvider, process_embeds
from .emoji import CustomEmoji, PartialEmoji, process_emoji, process_emoji_req_format
from .enums import (
    ActivityFlag,
    ActivityType,
    ApplicationFlags,
    AuditLogEventType,
    AutoArchiveDuration,
    ButtonStyle,
    ChannelFlags,
    ChannelType,
    CommandType,
    ComponentType,
    DefaultNotificationLevel,
    ExplicitContentFilterLevel,
    ForumLayoutType,
    IntegrationExpireBehaviour,
    Intents,
    InteractionPermissionTypes,
    InteractionType,
    InviteTargetType,
    MemberFlags,
    MentionType,
    MessageActivityType,
    MessageFlags,
    MessageType,
    MFALevel,
    NSFWLevel,
    OverwriteType,
    Permissions,
    PremiumTier,
    PremiumType,
    ScheduledEventPrivacyLevel,
    ScheduledEventStatus,
    ScheduledEventType,
    StagePrivacyLevel,
    Status,
    StickerFormatType,
    StickerTypes,
    SystemChannelFlags,
    TeamMembershipState,
    UserFlags,
    VerificationLevel,
    VideoQualityMode,
    WebSocketOPCode,
    ForumSortOrder,
)
from .file import File, open_file, UPLOADABLE_TYPE
from .guild import (
    AuditLog,
    AuditLogChange,
    AuditLogEntry,
    AuditLogHistory,
    BaseGuild,
    Guild,
    GuildBan,
    GuildIntegration,
    GuildPreview,
    GuildTemplate,
    GuildWelcome,
    GuildWelcomeChannel,
    GuildWidget,
    GuildWidgetSettings,
)
from .invite import Invite
from .message import (
    AllowedMentions,
    Attachment,
    BaseMessage,
    ChannelMention,
    Message,
    MessageActivity,
    MessageInteraction,
    MessageReference,
    process_allowed_mentions,
    process_message_payload,
    process_message_reference,
)
from .modal import InputText, Modal, ParagraphText, ShortText, TextStyles
from .reaction import Reaction, ReactionUsers
from .role import Role
from .scheduled_event import ScheduledEvent
from .snowflake import (
    Snowflake,
    Snowflake_Type,
    SnowflakeObject,
    to_optional_snowflake,
    to_snowflake,
    to_snowflake_list,
)
from .stage_instance import StageInstance
from .sticker import Sticker, StickerItem, StickerPack
from .team import Team, TeamMember
from .thread import ThreadList, ThreadMember, ThreadTag, DefaultReaction, process_thread_tag, process_default_reaction
from .timestamp import Timestamp, TimestampStyles
from .user import BaseUser, Member, User, ClientUser
from .voice_state import VoiceRegion, VoiceState

__all__ = (
    "ActionRow",
    "Activity",
    "ActivityAssets",
    "ActivityFlag",
    "ActivityParty",
    "ActivitySecrets",
    "ActivityTimestamps",
    "ActivityType",
    "AllowedMentions",
    "Application",
    "ApplicationCommandPermission",
    "ApplicationFlags",
    "Asset",
    "Attachment",
    "AuditLog",
    "AuditLogChange",
    "AuditLogEntry",
    "AuditLogEventType",
    "AuditLogHistory",
    "AutoArchiveDuration",
    "AutoModerationAction",
    "AutoModRule",
    "BaseChannel",
    "BaseComponent",
    "BaseGuild",
    "BaseMessage",
    "BaseSelectMenu",
    "BaseUser",
    "BrandColors",
    "BrandColours",
    "Button",
    "ButtonStyle",
    "ChannelFlags",
    "ChannelHistory",
    "ChannelMention",
    "ChannelSelectMenu",
    "ChannelType",
    "ClientUser",
    "Color",
    "COLOR_TYPES",
    "Colour",
    "CommandType",
    "ComponentType",
    "CustomEmoji",
    "DefaultNotificationLevel",
    "DefaultReaction",
    "DM",
    "DMChannel",
    "DMGroup",
    "Embed",
    "EmbedAttachment",
    "EmbedAuthor",
    "EmbedField",
    "EmbedFooter",
    "EmbedProvider",
    "ExplicitContentFilterLevel",
    "File",
    "FlatUIColors",
    "FlatUIColours",
    "ForumSortOrder",
    "ForumLayoutType",
    "get_components_ids",
    "Guild",
    "GuildBan",
    "GuildCategory",
    "GuildChannel",
    "GuildForum",
    "GuildForumPost",
    "GuildIntegration",
    "GuildMedia",
    "GuildNews",
    "GuildNewsThread",
    "GuildPreview",
    "GuildPrivateThread",
    "GuildPublicThread",
    "GuildStageVoice",
    "GuildTemplate",
    "GuildText",
    "GuildVoice",
    "GuildWelcome",
    "GuildWelcomeChannel",
    "GuildWidget",
    "GuildWidgetSettings",
    "InputText",
    "IntegrationExpireBehaviour",
    "Intents",
    "InteractionPermissionTypes",
    "InteractionType",
    "InteractiveComponent",
    "InvitableMixin",
    "Invite",
    "InviteTargetType",
    "MaterialColors",
    "MaterialColours",
    "Member",
    "MemberFlags",
    "MentionableSelectMenu",
    "MentionType",
    "Message",
    "MessageableMixin",
    "MessageActivity",
    "MessageActivityType",
    "MessageFlags",
    "MessageInteraction",
    "MessageReference",
    "MessageType",
    "MFALevel",
    "Modal",
    "NSFWLevel",
    "open_file",
    "OverwriteType",
    "ParagraphText",
    "PartialEmoji",
    "PermissionOverwrite",
    "Permissions",
    "PremiumTier",
    "PremiumType",
    "process_allowed_mentions",
    "process_color",
    "process_colour",
    "process_components",
    "process_default_reaction",
    "process_embeds",
    "process_emoji",
    "process_emoji_req_format",
    "process_message_payload",
    "process_message_reference",
    "process_permission_overwrites",
    "process_thread_tag",
    "Reaction",
    "ReactionUsers",
    "Role",
    "RoleColors",
    "RoleColours",
    "RoleSelectMenu",
    "ScheduledEvent",
    "ScheduledEventPrivacyLevel",
    "ScheduledEventStatus",
    "ScheduledEventType",
    "ShortText",
    "Snowflake",
    "Snowflake_Type",
    "SnowflakeObject",
    "spread_to_rows",
    "StageInstance",
    "StagePrivacyLevel",
    "Status",
    "Sticker",
    "StickerFormatType",
    "StickerItem",
    "StickerPack",
    "StickerTypes",
    "StringSelectMenu",
    "StringSelectOption",
    "SystemChannelFlags",
    "Team",
    "TeamMember",
    "TeamMembershipState",
    "TextStyles",
    "ThreadableMixin",
    "ThreadChannel",
    "ThreadList",
    "ThreadMember",
    "ThreadTag",
    "Timestamp",
    "TimestampStyles",
    "to_optional_snowflake",
    "to_snowflake",
    "to_snowflake_list",
    "TYPE_ALL_CHANNEL",
    "TYPE_CHANNEL_MAPPING",
    "TYPE_COMPONENT_MAPPING",
    "TYPE_DM_CHANNEL",
    "TYPE_GUILD_CHANNEL",
    "TYPE_MESSAGEABLE_CHANNEL",
    "TYPE_THREAD_CHANNEL",
    "TYPE_VOICE_CHANNEL",
    "UPLOADABLE_TYPE",
    "User",
    "UserFlags",
    "UserSelectMenu",
    "VerificationLevel",
    "VideoQualityMode",
    "VoiceRegion",
    "VoiceState",
    "Webhook",
    "WebhookMixin",
    "WebhookTypes",
    "WebSocketOPCode",
)
from .webhooks import Webhook, WebhookTypes



from __future__ import annotations

import asyncio
import typing
from typing import (
    Annotated,
    Awaitable,
    Callable,
    Coroutine,
    Optional,
    Tuple,
    Any,
    TYPE_CHECKING,
    TypeVar,
)

import attrs

from interactions.client.const import MISSING, AsyncCallable
from interactions.client.errors import CommandOnCooldown, CommandCheckFailure, MaxConcurrencyReached
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.attr_utils import docs
from interactions.client.utils.misc_utils import get_parameters, get_object_name, maybe_coroutine
from interactions.client.utils.serializer import no_export_meta
from interactions.models.internal.callback import CallbackObject
from interactions.models.internal.cooldowns import Cooldown, Buckets, MaxConcurrency, CooldownSystem
from interactions.models.internal.protocols import Converter

if TYPE_CHECKING:
    from interactions.models.internal.extension import Extension
    from interactions.models.internal.context import BaseContext

__all__ = ("BaseCommand", "check", "cooldown", "max_concurrency")


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class BaseCommand(DictSerializationMixin, CallbackObject):
    """
    An object all commands inherit from. Outlines the basic structure of a command, and handles checks.

    Attributes:
        extension: The extension this command belongs to.
        enabled: Whether this command is enabled
        checks: Any checks that must be run before this command can be run
        callback: The coroutine to be called for this command
        error_callback: The coroutine to be called when an error occurs
        pre_run_callback: A coroutine to be called before this command is run **but** after the checks
        post_run_callback: A coroutine to be called after this command has run

    """

    extension: "Optional[Extension]" = attrs.field(
        repr=False,
        default=None,
        metadata=docs("The extension this command belongs to") | no_export_meta,
    )

    enabled: bool = attrs.field(
        repr=False, default=True, metadata=docs("Whether this can be run at all") | no_export_meta
    )
    checks: list = attrs.field(
        repr=False,
        factory=list,
        metadata=docs("Any checks that must be *checked* before the command can run") | no_export_meta,
    )
    cooldown: Cooldown = attrs.field(
        repr=False,
        default=MISSING,
        metadata=docs("An optional cooldown to apply to the command") | no_export_meta,
    )
    max_concurrency: MaxConcurrency = attrs.field(
        default=MISSING,
        metadata=docs("An optional maximum number of concurrent instances to apply to the command") | no_export_meta,
    )

    callback: Callable[..., Coroutine] = attrs.field(
        repr=False,
        default=None,
        metadata=docs("The coroutine to be called for this command") | no_export_meta,
    )
    error_callback: Callable[..., Coroutine] = attrs.field(
        repr=False,
        default=None,
        metadata=no_export_meta | docs("The coroutine to be called when an error occurs"),
    )
    pre_run_callback: Callable[..., Coroutine] = attrs.field(
        default=None,
        metadata=no_export_meta
        | docs("The coroutine to be called before the command is executed, **but** after the checks"),
    )
    post_run_callback: Callable[..., Coroutine] = attrs.field(
        repr=False,
        default=None,
        metadata=no_export_meta | docs("The coroutine to be called after the command has executed"),
    )

    def __attrs_post_init__(self) -> None:
        if self.callback is not None:
            if hasattr(self.callback, "checks"):
                self.checks += self.callback.checks
            if hasattr(self.callback, "cooldown"):
                self.cooldown = self.callback.cooldown
            if hasattr(self.callback, "max_concurrency"):
                self.max_concurrency = self.callback.max_concurrency

    def __hash__(self) -> int:
        return id(self)

    async def __call__(self, context: "BaseContext", *args, **kwargs) -> None:
        """
        Calls this command.

        Args:
            context: The context of this command
            args: Any
            kwargs: Any

        """
        # signals if a semaphore has been acquired, for exception handling
        # if present assume one will be acquired
        max_conc_acquired = self.max_concurrency is not MISSING

        try:
            if await self._can_run(context):
                if self.pre_run_callback is not None:
                    await self.call_with_binding(self.pre_run_callback, context, *args, **kwargs)

                if self.extension is not None and self.extension.extension_prerun:
                    for prerun in self.extension.extension_prerun:
                        await prerun(context, *args, **kwargs)

                await self.call_callback(self.callback, context)

                if self.post_run_callback is not None:
                    await self.call_with_binding(self.post_run_callback, context, *args, **kwargs)

                if self.extension is not None and self.extension.extension_postrun:
                    for postrun in self.extension.extension_postrun:
                        await postrun(context, *args, **kwargs)

        except Exception as e:
            # if a MaxConcurrencyReached-exception is raised a connection was never acquired
            max_conc_acquired = not isinstance(e, MaxConcurrencyReached)

            if self.error_callback:
                await self.error_callback(e, context, *args, **kwargs)
            elif self.extension and self.extension.extension_error:
                await self.extension.extension_error(e, context, *args, **kwargs)
            else:
                raise
        finally:
            if self.max_concurrency is not MISSING and max_conc_acquired:
                await self.max_concurrency.release(context)

    @staticmethod
    def _get_converter_function(anno: type[Converter] | Converter, name: str) -> Callable[[BaseContext, str], Any]:
        num_params = len(get_parameters(anno.convert))

        # if we have three parameters for the function, it's likely it has a self parameter
        # so we need to get rid of it by initing - typehinting hates this, btw!
        # the below line will error out if we aren't supposed to init it, so that works out
        try:
            actual_anno: Converter = anno() if num_params == 3 else anno  # type: ignore
        except TypeError:
            raise ValueError(
                f"{get_object_name(anno)} for {name} is invalid: converters must have exactly 2 arguments."
            ) from None

        # we can only get to this point while having three params if we successfully inited
        if num_params == 3:
            num_params -= 1

        if num_params != 2:
            raise ValueError(
                f"{get_object_name(anno)} for {name} is invalid: converters must have exactly 2 arguments."
            )

        return actual_anno.convert

    async def try_convert(self, converter: Optional[Callable], context: "BaseContext", value: Any) -> Any:
        if converter is None:
            return value
        return await maybe_coroutine(converter, context, value)

    def param_config(self, annotation: Any, name: str) -> Tuple[Callable, Optional[dict]]:
        # This thing is complicated. i.py-annotations can either be annotated directly, or they can be annotated with Annotated[str, CMD_*]
        # This helper function handles both cases, and returns a tuple of the converter and its config (if any)
        if annotation is None:
            return None
        if typing.get_origin(annotation) is Annotated and (args := typing.get_args(annotation)):
            for ann in args:
                v = getattr(ann, name, None)
                if v is not None:
                    return (ann, v)
        return (annotation, getattr(annotation, name, None))

    async def call_callback(self, callback: Callable, context: "BaseContext") -> None:
        await self.call_with_binding(callback, context, **context.kwargs)  # type: ignore

    async def _can_run(self, context: "BaseContext") -> bool:
        """
        Determines if this command can be run.

        Args:
            context: The context of the command

        """
        max_conc_acquired = False  # signals if a semaphore has been acquired, for exception handling

        try:
            if not self.enabled:
                return False

            for _c in self.checks:
                if not await _c(context):
                    raise CommandCheckFailure(self, _c, context)

            if self.extension and self.extension.extension_checks:
                for _c in self.extension.extension_checks:
                    if not await _c(context):
                        raise CommandCheckFailure(self, _c, context)

            if self.max_concurrency is not MISSING and not await self.max_concurrency.acquire(context):
                raise MaxConcurrencyReached(self, self.max_concurrency)

            if self.cooldown is not MISSING and not await self.cooldown.acquire_token(context):
                raise CommandOnCooldown(self, await self.cooldown.get_cooldown(context))

            return True

        except Exception:
            if max_conc_acquired:
                await self.max_concurrency.release(context)
            raise

    def add_check(self, check: Callable[..., Awaitable[bool]]) -> None:
        """Adds a check into the command."""
        self.checks.append(check)

    def error(self, call: Callable[..., Coroutine]) -> Callable[..., Coroutine]:
        """A decorator to declare a coroutine as one that will be run upon an error."""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("Error handler must be coroutine")
        self.error_callback = call
        return call

    def pre_run(self, call: Callable[..., Coroutine]) -> Callable[..., Coroutine]:
        """A decorator to declare a coroutine as one that will be run before the command."""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("pre_run must be coroutine")
        self.pre_run_callback = call
        return call

    def post_run(self, call: Callable[..., Coroutine]) -> Callable[..., Coroutine]:
        """A decorator to declare a coroutine as one that will be run after the command has."""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("post_run must be coroutine")
        self.post_run_callback = call
        return call


CommandT = TypeVar("CommandT", BaseCommand, AsyncCallable)


def check(check: Callable[..., Awaitable[bool]]) -> Callable[[CommandT], CommandT]:
    """
    Add a check to a command.

    Args:
        check: A coroutine as a check for this command

    """

    def wrapper(coro: CommandT) -> CommandT:
        if isinstance(coro, BaseCommand):
            coro.checks.append(check)
            return coro
        if not hasattr(coro, "checks"):
            coro.checks = []
        coro.checks.append(check)
        return coro

    return wrapper


def cooldown(
    bucket: Buckets, rate: int, interval: float, cooldown_system: typing.Type[CooldownSystem] | None = None
) -> Callable[[CommandT], CommandT]:
    """
    Add a cooldown to a command.

    Args:
        bucket: The bucket used to track cooldowns
        rate: How many commands may be ran per interval
        interval: How many seconds to wait for a cooldown
        cooldown_system: The cooldown system to use
    """

    def wrapper(coro: CommandT) -> CommandT:
        cooldown_obj = Cooldown(bucket, rate, interval, cooldown_system=cooldown_system)

        coro.cooldown = cooldown_obj

        return coro

    return wrapper


def max_concurrency(bucket: Buckets, concurrent: int) -> Callable[[CommandT], CommandT]:
    """
    Add a maximum number of concurrent instances to the command.

    Args:
        bucket: The bucket to enforce the maximum within
        concurrent: The maximum number of concurrent instances to allow

    """

    def wrapper(coro: CommandT) -> CommandT:
        max_conc = MaxConcurrency(concurrent, bucket)

        coro.max_concurrency = max_conc

        return coro

    return wrapper



import datetime

from typing import TYPE_CHECKING, Any, Optional, Union, Iterable, Sequence
from typing_extensions import Self


from interactions import (
    BaseContext,
    Permissions,
    Message,
    SlashContext,
    Client,
    Typing,
    Embed,
    BaseComponent,
    UPLOADABLE_TYPE,
    Snowflake_Type,
    Sticker,
    AllowedMentions,
    MessageReference,
    MessageFlags,
    to_snowflake,
    Attachment,
    process_message_payload,
)
from interactions.client.mixins.send import SendMixin
from interactions.ext import prefixed_commands as prefixed

if TYPE_CHECKING:
    from .hybrid_slash import HybridSlashCommand

__all__ = ("HybridContext",)


class DeferTyping:
    def __init__(self, ctx: "HybridContext", ephermal: bool) -> None:
        self.ctx = ctx
        self.ephermal = ephermal

    async def __aenter__(self) -> None:
        await self.ctx.defer(ephemeral=self.ephermal)

    async def __aexit__(self, *_) -> None:
        pass


class HybridContext(BaseContext, SendMixin):
    prefix: str
    "The prefix used to invoke this command."

    app_permissions: Permissions
    """The permissions available to this context"""

    deferred: bool
    """Whether the context has been deferred."""
    responded: bool
    """Whether the context has been responded to."""
    ephemeral: bool
    """Whether the context response is ephemeral."""

    _command_name: str
    """The command name."""
    _message: Message | None

    args: list[Any]
    """The arguments passed to the command."""
    kwargs: dict[str, Any]
    """The keyword arguments passed to the command."""

    __attachment_index__: int

    _slash_ctx: SlashContext | None
    _prefixed_ctx: prefixed.PrefixedContext | None

    def __init__(self, client: Client):
        super().__init__(client)
        self.prefix = ""
        self.app_permissions = Permissions(0)
        self.deferred = False
        self.responded = False
        self.ephemeral = False
        self._command_name = ""
        self.args = []
        self.kwargs = {}
        self._message = None
        self.__attachment_index__ = 0
        self._slash_ctx = None
        self._prefixed_ctx = None

    @classmethod
    def from_dict(cls, client: Client, payload: dict) -> None:
        # this doesn't mean anything, so just implement it to make abc happy
        raise NotImplementedError

    @classmethod
    def from_slash_context(cls, ctx: SlashContext) -> Self:
        self = cls(ctx.client)
        self.guild_id = ctx.guild_id
        self.channel_id = ctx.channel_id
        self.author_id = ctx.author_id
        self.message_id = ctx.message_id
        self.prefix = "/"
        self.app_permissions = ctx.app_permissions
        self.deferred = ctx.deferred
        self.responded = ctx.responded
        self.ephemeral = ctx.ephemeral
        self._command_name = ctx._command_name
        self.args = ctx.args
        self.kwargs = ctx.kwargs
        self._slash_ctx = ctx
        return self

    @classmethod
    def from_prefixed_context(cls, ctx: prefixed.PrefixedContext) -> Self:
        # this is a "best guess" on what the permissions are
        # this may or may not be totally accurate
        if hasattr(ctx.channel, "permissions_for"):
            app_permissions = ctx.channel.permissions_for(ctx.guild.me)  # type: ignore
        elif ctx.channel.type in {10, 11, 12}:  # it's a thread
            app_permissions = ctx.channel.parent_channel.permissions_for(ctx.guild.me)  # type: ignore
        else:
            app_permissions = Permissions(0)

        self = cls(ctx.client)
        self.guild_id = ctx.guild_id
        self.channel_id = ctx.channel_id
        self.author_id = ctx.author_id
        self.message_id = ctx.message_id
        self._message = ctx.message
        self.prefix = ctx.prefix
        self.app_permissions = app_permissions
        self._command_name = ctx.command.qualified_name
        self.args = ctx.args
        self._prefixed_ctx = ctx
        return self

    @property
    def inner_context(self) -> SlashContext | prefixed.PrefixedContext:
        """The inner context that this hybrid context is wrapping."""
        return self._slash_ctx or self._prefixed_ctx  # type: ignore

    @property
    def command(self) -> "HybridSlashCommand":
        return self.client._interaction_lookup[self._command_name]

    @property
    def expires_at(self) -> Optional[datetime.datetime]:
        """The time at which the interaction expires."""
        if not self._slash_ctx:
            return None

        if self.responded:
            return self._slash_ctx.id.created_at + datetime.timedelta(minutes=15)
        return self._slash_ctx.id.created_at + datetime.timedelta(seconds=3)

    @property
    def expired(self) -> bool:
        """Whether the interaction has expired."""
        return datetime.datetime.utcnow() > self.expires_at if self._slash_ctx else False

    @property
    def deferred_ephemeral(self) -> bool:
        """Whether the interaction has been deferred ephemerally."""
        return self.deferred and self.ephemeral

    @property
    def message(self) -> Message | None:
        """The message that invoked this context."""
        return self._message or self.client.cache.get_message(self.channel_id, self.message_id)

    @property
    def typing(self) -> Typing | DeferTyping:
        """A context manager to send a _typing/defer state to a given channel as long as long as the wrapped operation takes."""
        if self._slash_ctx:
            return DeferTyping(self._slash_ctx, self.ephemeral)
        return self.channel.typing

    async def defer(self, ephemeral: bool = False) -> None:
        """
        Either defers the response (if used in an interaction) or triggers a typing indicator for 10 seconds (if used for messages).

        Args:
            ephemeral: Should the response be ephemeral? Only applies to responses for interactions.

        """
        if self._slash_ctx:
            await self._slash_ctx.defer(ephemeral=ephemeral)
        else:
            await self.channel.trigger_typing()

        self.deferred = True

    async def reply(
        self,
        content: Optional[str] = None,
        embeds: Optional[
            Union[
                Iterable[Union[Embed, dict]],
                Union[Embed, dict],
            ]
        ] = None,
        embed: Optional[Union[Embed, dict]] = None,
        **kwargs,
    ) -> "Message":
        """
        Reply to this message, takes all the same attributes as `send`.

        For interactions, this functions the same as `send`.
        """
        kwargs = locals()
        kwargs.pop("self")
        extra_kwargs = kwargs.pop("kwargs")
        kwargs |= extra_kwargs

        if self._slash_ctx:
            result = await self.send(**kwargs)
        else:
            kwargs.pop("ephemeral", None)
            result = await self._prefixed_ctx.reply(**kwargs)

        self.responded = True
        return result

    async def _send_http_request(
        self,
        message_payload: dict,
        files: Iterable["UPLOADABLE_TYPE"] | None = None,
    ) -> dict:
        if self._slash_ctx:
            return await self._slash_ctx._send_http_request(message_payload, files)
        return await self._prefixed_ctx._send_http_request(message_payload, files)

    async def send(
        self,
        content: Optional[str] = None,
        *,
        embeds: Optional[
            Union[
                Iterable[Union["Embed", dict]],
                Union["Embed", dict],
            ]
        ] = None,
        embed: Optional[Union["Embed", dict]] = None,
        components: Optional[
            Union[
                Iterable[Iterable[Union["BaseComponent", dict]]],
                Iterable[Union["BaseComponent", dict]],
                "BaseComponent",
                dict,
            ]
        ] = None,
        stickers: Optional[
            Union[
                Iterable[Union["Sticker", "Snowflake_Type"]],
                "Sticker",
                "Snowflake_Type",
            ]
        ] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        reply_to: Optional[Union["MessageReference", "Message", dict, "Snowflake_Type"]] = None,
        files: Optional[Union["UPLOADABLE_TYPE", Iterable["UPLOADABLE_TYPE"]]] = None,
        file: Optional["UPLOADABLE_TYPE"] = None,
        tts: bool = False,
        suppress_embeds: bool = False,
        silent: bool = False,
        flags: Optional[Union[int, "MessageFlags"]] = None,
        delete_after: Optional[float] = None,
        ephemeral: bool = False,
        **kwargs: Any,
    ) -> "Message":
        """
        Send a message.

        Args:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            embed: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            stickers: IDs of up to 3 stickers in the server to send in the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            files: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            file: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            tts: Should this message use Text To Speech.
            suppress_embeds: Should embeds be suppressed on this send
            silent: Should this message be sent without triggering a notification.
            flags: Message flags to apply.
            delete_after: Delete message after this many seconds.
            ephemeral: Should this message be sent as ephemeral (hidden) - only works with interactions

        Returns:
            New message object that was sent.
        """
        flags = MessageFlags(flags or 0)
        if ephemeral and self._slash_ctx:
            flags |= MessageFlags.EPHEMERAL
            self.ephemeral = True
        if suppress_embeds:
            flags |= MessageFlags.SUPPRESS_EMBEDS
        if silent:
            flags |= MessageFlags.SILENT

        return await super().send(
            content=content,
            embeds=embeds,
            embed=embed,
            components=components,
            stickers=stickers,
            allowed_mentions=allowed_mentions,
            reply_to=reply_to,
            files=files,
            file=file,
            tts=tts,
            flags=flags,
            delete_after=delete_after,
            pass_self_into_delete=bool(self._slash_ctx),
            **kwargs,
        )

    async def delete(self, message: "Snowflake_Type") -> None:
        """
        Delete a message sent in response to this context. Must be in the same channel as the context.

        Args:
            message: The message to delete
        """
        if self._slash_ctx:
            return await self._slash_ctx.delete(message)
        await self.client.http.delete_message(self.channel_id, to_snowflake(message))

    async def edit(
        self,
        message: "Snowflake_Type",
        *,
        content: Optional[str] = None,
        embeds: Optional[
            Union[
                Iterable[Union["Embed", dict]],
                Union["Embed", dict],
            ]
        ] = None,
        embed: Optional[Union["Embed", dict]] = None,
        components: Optional[
            Union[
                Iterable[Iterable[Union["BaseComponent", dict]]],
                Iterable[Union["BaseComponent", dict]],
                "BaseComponent",
                dict,
            ]
        ] = None,
        attachments: Optional[Sequence[Attachment | dict]] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        files: Optional[Union["UPLOADABLE_TYPE", Iterable["UPLOADABLE_TYPE"]]] = None,
        file: Optional["UPLOADABLE_TYPE"] = None,
        tts: bool = False,
    ) -> "Message":
        if self._slash_ctx:
            return await self._slash_ctx.edit(
                message,
                content=content,
                embeds=embeds,
                embed=embed,
                components=components,
                attachments=attachments,
                allowed_mentions=allowed_mentions,
                files=files,
                file=file,
                tts=tts,
            )

        message_payload = process_message_payload(
            content=content,
            embeds=embeds or embed,
            components=components,
            allowed_mentions=allowed_mentions,
            attachments=attachments,
            tts=tts,
        )
        if file:
            files = [file, *files] if files else [file]

        message_data = await self.client.http.edit_message(
            message_payload, self.channel_id, to_snowflake(message), files=files
        )
        if message_data:
            return self.client.cache.place_message_data(message_data)



import functools
import inspect
import typing
from collections import deque
from types import NoneType, UnionType
from typing import (
    Optional,
    Any,
    Callable,
    Annotated,
    Literal,
    Union,
    TYPE_CHECKING,
    Type,
    TypeGuard,
)

import attrs
from typing_extensions import Self

from interactions.client.const import MISSING, T
from interactions.client.errors import BadArgument
from interactions.client.utils.input_utils import _quotes
from interactions.client.utils.misc_utils import get_object_name, maybe_coroutine
from interactions.models.internal.command import BaseCommand
from interactions.models.internal.converters import (
    _LiteralConverter,
    NoArgumentConverter,
    Greedy,
    CONSUME_REST_MARKER,
    MODEL_TO_CONVERTER,
)
from interactions.models.internal.protocols import Converter
from ...client.utils.attr_utils import docs

if TYPE_CHECKING:
    from .context import PrefixedContext

__all__ = (
    "PrefixedCommandParameter",
    "PrefixedCommand",
    "prefixed_command",
)


_STARTING_QUOTES = frozenset(_quotes.keys())


class PrefixedCommandParameter:
    """
    An object representing parameters in a prefixed command.

    This class should not be instantiated directly.
    """

    __slots__ = (
        "name",
        "default",
        "type",
        "kind",
        "converters",
        "greedy",
        "union",
        "variable",
        "consume_rest",
        "consume_rest_class",
        "no_argument",
    )

    name: str
    "The name of the parameter."
    default: Any
    "The default value of the parameter."
    type: Type
    "The type of the parameter."
    kind: inspect._ParameterKind
    """The kind of parameter this is as related to the function."""
    converters: list[Callable[["PrefixedContext", str], Any]]
    "A list of the converter functions for the parameter that convert to its type."
    greedy: bool
    "Is the parameter greedy?"
    union: bool
    "Is the parameter a union?"
    variable: bool
    "Was the parameter marked as a variable argument?"
    consume_rest: bool
    "Was the parameter marked to consume the rest of the input?"
    no_argument: bool
    "Does this parameter have a converter that subclasses `NoArgumentConverter`?"

    def __init__(
        self,
        name: str,
        default: Any = MISSING,
        type: Type | None = None,
        kind: inspect._ParameterKind = inspect._ParameterKind.POSITIONAL_OR_KEYWORD,
        converters: Optional[list[Callable[["PrefixedContext", str], Any]]] = None,
        greedy: bool = False,
        union: bool = False,
        variable: bool = False,
        consume_rest: bool = False,
        no_argument: bool = False,
    ) -> None:
        self.name = name
        self.default = default
        self.type = type
        self.kind = kind
        self.converters = converters or []
        self.greedy = greedy
        self.union = union
        self.variable = variable
        self.consume_rest = consume_rest
        self.no_argument = no_argument

    @classmethod
    def from_param(cls, param: inspect.Parameter) -> Self:
        return cls(
            param.name,
            param.default if param.default is not param.empty else MISSING,
            param.annotation,
            param.kind,
        )

    def __repr__(self) -> str:
        return f"<PrefixedCommandParameter name={self.name!r}>"

    @property
    def optional(self) -> bool:
        """Is this parameter optional?"""
        return self.default != MISSING


class _PrefixedArgsIterator:
    """
    An iterator over the arguments of a prefixed command.

    Has functions to control the iteration.
    """

    __slots__ = ("args", "index", "length")

    def __init__(self, args: tuple[str, ...]) -> None:
        self.args = args
        self.index = 0
        self.length = len(self.args)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> str:
        if self.index >= self.length:
            raise StopIteration

        result = self.args[self.index]
        self.index += 1
        return self._remove_quotes(result)

    def _remove_quotes(self, arg: str) -> str:
        # this removes quotes from the arguments themselves
        return arg[1:-1] if arg[0] in _STARTING_QUOTES else arg

    def _finish_args(self) -> tuple[str]:
        result = self.args[self.index - 1 :]
        self.index = self.length
        return result

    def get_rest_of_args(self) -> tuple[str]:
        return tuple(self._remove_quotes(r) for r in self._finish_args())

    def consume_rest(self) -> str:
        return " ".join(self._finish_args())

    def back(self, count: int = 1) -> None:
        self.index -= count

    def reset(self) -> None:
        self.index = 0

    @property
    def finished(self) -> bool:
        return self.index >= self.length


def _check_for_no_arg(anno: Any) -> TypeGuard[NoArgumentConverter]:
    return isinstance(anno, NoArgumentConverter) or (inspect.isclass(anno) and issubclass(anno, NoArgumentConverter))


def _convert_to_bool(argument: str) -> bool:
    lowered = argument.lower()
    if lowered in {"yes", "y", "true", "t", "1", "enable", "on"}:
        return True
    if lowered in {"no", "n", "false", "f", "0", "disable", "off"}:
        return False
    raise BadArgument(f"{argument} is not a recognised boolean option.")


def _get_from_anno_type(anno: Annotated) -> Any:
    """
    Handles dealing with Annotated annotations, getting their (first) type annotation.

    This allows correct type hinting with, say, Converters, for example.
    """
    # this is treated how it usually is during runtime
    # the first argument is ignored and the rest is treated as is

    return typing.get_args(anno)[1]


def _get_converter(anno: type, name: str) -> Callable[["PrefixedContext", str], Any]:  # type: ignore
    if typing.get_origin(anno) == Annotated:
        anno = _get_from_anno_type(anno)

    if isinstance(anno, Converter):
        return BaseCommand._get_converter_function(anno, name)
    if converter := MODEL_TO_CONVERTER.get(anno, None):
        return BaseCommand._get_converter_function(converter, name)
    if typing.get_origin(anno) is Literal:
        literals = typing.get_args(anno)
        return _LiteralConverter(literals).convert
    if inspect.isroutine(anno):
        num_params = len(inspect.signature(anno).parameters.values())
        match num_params:
            case 2:
                return anno
            case 1:

                async def _one_function_cmd(ctx, arg) -> Any:
                    return await maybe_coroutine(anno, arg)

                return _one_function_cmd
            case 0:
                ValueError(f"{get_object_name(anno)} for {name} has 0 arguments, which is unsupported.")
            case _:
                ValueError(f"{get_object_name(anno)} for {name} has more than 2 arguments, which is unsupported.")
    elif anno == bool:
        return lambda ctx, arg: _convert_to_bool(arg)
    elif anno == inspect._empty:
        return lambda ctx, arg: str(arg)
    else:
        return lambda ctx, arg: anno(arg)


def _greedy_parse(greedy: Greedy, param: inspect.Parameter) -> Any:
    default = param.default

    if param.kind in {param.KEYWORD_ONLY, param.VAR_POSITIONAL}:
        raise ValueError("Greedy[...] cannot be a variable or keyword-only argument.")

    arg = typing.get_args(greedy)[0]

    if typing.get_origin(arg) == Annotated:
        arg = _get_from_anno_type(arg)

    if typing.get_origin(arg) in {Union, UnionType}:
        args = typing.get_args(arg)

        if len(args) > 2 or NoneType not in args:
            raise ValueError(f"Greedy[{arg!r}] is invalid.")

        arg = args[0]
        default = None

    if arg in {NoneType, str, Greedy, Union, UnionType}:
        raise ValueError(f"Greedy[{get_object_name(arg)}] is invalid.")

    return arg, default


async def _convert(param: PrefixedCommandParameter, ctx: "PrefixedContext", arg: str) -> tuple[Any, bool]:
    converted = MISSING
    for converter in param.converters:
        try:
            converted = await maybe_coroutine(converter, ctx, arg)
            break
        except Exception as e:
            if not param.union and not param.optional:
                if isinstance(e, BadArgument):
                    raise
                raise BadArgument(str(e)) from e

    used_default = False
    if converted == MISSING:
        if param.optional:
            converted = param.default
            used_default = True
        else:
            union_types = typing.get_args(param.type)
            union_names = tuple(get_object_name(t) for t in union_types)
            union_types_str = ", ".join(union_names[:-1]) + f", or {union_names[-1]}"
            raise BadArgument(f'Could not convert "{arg}" into {union_types_str}.')

    if param.no_argument:
        # tells converter not to eat the current argument
        used_default = True

    return converted, used_default


async def _greedy_convert(
    param: PrefixedCommandParameter, ctx: "PrefixedContext", args: _PrefixedArgsIterator
) -> tuple[list[Any] | Any, bool]:
    args.back()
    broke_off = False
    greedy_args = []

    for arg in args:
        try:
            greedy_arg, used_default = await _convert(param, ctx, arg)

            if used_default:
                raise BadArgument

            greedy_args.append(greedy_arg)
        except BadArgument:
            broke_off = True
            break

    if not greedy_args:
        if param.default:
            greedy_args = param.default  # im sorry, typehinters
        else:
            raise BadArgument(f"Failed to find any arguments for {param.type!r}.")

    return greedy_args, broke_off


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class PrefixedCommand(BaseCommand):
    """Represents a command based off a message, usually denoted with a prefix."""

    name: str = attrs.field(repr=False, metadata=docs("The name of the command."))
    parameters: list[PrefixedCommandParameter] = attrs.field(
        repr=False, metadata=docs("The parameters of the command."), factory=list
    )
    aliases: list[str] = attrs.field(
        metadata=docs("The list of aliases the command can be invoked under."),
        factory=list,
    )
    hidden: bool = attrs.field(
        metadata=docs("If `True`, help commands should not show this in the help output (unless toggled to do so)."),
        default=False,
    )
    ignore_extra: bool = attrs.field(
        metadata=docs(
            "If `True`, ignores extraneous strings passed to a command if all its requirements are met (e.g. ?foo a b c"
            " when only expecting a and b). Otherwise, an error is raised. Defaults to True."
        ),
        default=True,
    )
    help: Optional[str] = attrs.field(repr=False, metadata=docs("The long help text for the command."), default=None)
    brief: Optional[str] = attrs.field(repr=False, metadata=docs("The short help text for the command."), default=None)
    parent: Optional["PrefixedCommand"] = attrs.field(
        repr=False, metadata=docs("The parent command, if applicable."), default=None
    )
    subcommands: dict[str, "PrefixedCommand"] = attrs.field(
        repr=False, metadata=docs("A dict of all subcommands for the command."), factory=dict
    )
    _usage: Optional[str] = attrs.field(repr=False, default=None)
    _inspect_signature: Optional[inspect.Signature] = attrs.field(repr=False, default=None)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()  # we want checks to work

        # we have to do this afterwards as these rely on the callback
        # and its own value, which is impossible to get with attrs
        # methods, i think

        if self.help:
            self.help = inspect.cleandoc(self.help)
        else:
            self.help = inspect.getdoc(self.callback)
            if isinstance(self.help, bytes):
                self.help = self.help.decode("utf-8")

        if self.brief is None:
            self.brief = self.help.splitlines()[0] if self.help is not None else None

    @property
    def usage(self) -> str:
        """
        A string displaying how the command can be used.

        If no string is set, it will default to the command's signature.
        Useful for help commands.
        """
        return self._usage or self.signature

    @usage.setter
    def usage(self, usage: str) -> None:
        self._usage = usage

    @property
    def qualified_name(self) -> str:
        """Returns the full qualified name of this command."""
        name_deq = deque()
        command = self

        while command.parent is not None:
            name_deq.appendleft(command.name)
            command = command.parent

        name_deq.appendleft(command.name)
        return " ".join(name_deq)

    @property
    def all_subcommands(self) -> frozenset[Self]:
        """Returns all unique subcommands underneath this command."""
        return frozenset(self.subcommands.values())

    @property
    def signature(self) -> str:
        """Returns a POSIX-like signature useful for help command output."""
        if not self.parameters:
            return ""

        results = []

        for param in self.parameters:
            anno = param.type
            name = param.name

            if typing.get_origin(anno) == Annotated:
                anno = typing.get_args(anno)[1]

            if not param.greedy and param.union:
                union_args = typing.get_args(anno)
                if len(union_args) == 2 and param.optional:
                    anno = union_args[0]

            if typing.get_origin(anno) is Literal:
                # it's better to list the values it can be than display the variable name itself
                name = "|".join(f'"{v}"' if isinstance(v, str) else str(v) for v in typing.get_args(anno))

            # we need to do a lot of manipulations with the signature
            # string, so using a deque as a string builder makes sense for performance
            result_builder: deque[str] = deque()

            if param.optional and param.default is not None:
                # it would be weird making it look like name=None
                result_builder.append(f"{name}={param.default}")
            else:
                result_builder.append(name)

            if param.variable:
                # this is inside the brackets
                result_builder.append("...")

            # surround the result with brackets
            if param.optional:
                result_builder.appendleft("[")
                result_builder.append("]")
            else:
                result_builder.appendleft("<")
                result_builder.append(">")

            if param.greedy:
                # this is outside the brackets, making it differentiable from
                # a variable argument
                result_builder.append("...")

            results.append("".join(result_builder))

        return " ".join(results)

    @property
    def is_subcommand(self) -> bool:
        """Return whether this command is a subcommand or not."""
        return bool(self.parent)

    def _parse_parameters(self) -> None:  # noqa: C901
        """
        Parses the parameters that this command has into a form i.py can use.

        This is purposely separated like this to allow "lazy parsing" - parsing
        as the command is added to a bot rather than being parsed immediately.
        This allows variables like "self" to be filtered out, and is useful for
        potential future additions.
        """
        # clear out old parameters just in case
        self.parameters = []

        if not self._inspect_signature:
            # we don't care about the ctx or self variables
            if self.has_binding:
                callback = functools.partial(self.callback, None, None)
            else:
                callback = functools.partial(self.callback, None)

            self._inspect_signature = inspect.signature(callback)

        params = self._inspect_signature.parameters

        # this is used by keyword-only and variable args to make sure there isn't more than one of either
        # mind you, we also don't want one keyword-only and one variable arg either
        finished_params = False

        for name, param in params.items():
            if finished_params:
                raise ValueError("Cannot have multiple keyword-only or variable arguments.")

            cmd_param = PrefixedCommandParameter.from_param(param)
            anno = param.annotation

            # this is ugly, ik
            if typing.get_origin(anno) == Annotated and typing.get_args(anno)[1] is CONSUME_REST_MARKER:
                cmd_param.consume_rest = True
                finished_params = True
                anno = typing.get_args(anno)[0]

                if anno == T:
                    # someone forgot to typehint
                    anno = inspect._empty

            if typing.get_origin(anno) == Annotated:
                anno = _get_from_anno_type(anno)

            if typing.get_origin(anno) == Greedy:
                if finished_params:
                    raise ValueError("Consume rest arguments cannot be Greedy.")

                anno, default = _greedy_parse(anno, param)

                if default is not param.empty:
                    cmd_param.default = default
                cmd_param.greedy = True

            if typing.get_origin(anno) == tuple:
                if cmd_param.optional:
                    # there's a lot of parser ambiguities here, so i'd rather not
                    raise ValueError("Variable arguments cannot have default values or be Optional.")
                cmd_param.variable = True
                finished_params = True

                # use empty if the typehint is just "tuple"
                anno = typing.get_args(anno)[0] if typing.get_args(anno) else inspect._empty

            if typing.get_origin(anno) in {Union, UnionType}:
                cmd_param.union = True
                for arg in typing.get_args(anno):
                    if _check_for_no_arg(anno):
                        cmd_param.no_argument = True

                    if arg != NoneType:
                        converter = _get_converter(arg, name)
                        cmd_param.converters.append(converter)
                    elif not cmd_param.optional:  # d.py-like behavior
                        cmd_param.default = None
            else:
                if _check_for_no_arg(anno):
                    cmd_param.no_argument = True

                converter = _get_converter(anno, name)
                cmd_param.converters.append(converter)

            if not finished_params:
                match param.kind:
                    case param.KEYWORD_ONLY:
                        if cmd_param.greedy:
                            raise ValueError("Consume rest arguments cannot be Greedy.")

                        cmd_param.consume_rest = True
                        finished_params = True
                    case param.VAR_POSITIONAL:
                        if cmd_param.optional:
                            # there's a lot of parser ambiguities here, so i'd rather not
                            raise ValueError("Variable arguments cannot have default values or be Optional.")
                        if cmd_param.greedy:
                            raise ValueError("Variable arguments cannot be Greedy.")

                        cmd_param.variable = True
                        finished_params = True

            self.parameters.append(cmd_param)

        # we need to deal with subcommands too
        for command in self.all_subcommands:
            command._parse_parameters()

    def add_command(self, cmd: Self) -> None:
        """
        Adds a command as a subcommand to this command.

        Args:
            cmd: The command to add
        """
        cmd.parent = self  # just so we know this is a subcommand

        if self.subcommands.get(cmd.name):
            raise ValueError(
                f"Duplicate command! Multiple commands share the name/alias: {self.qualified_name} {cmd.name}."
            )
        self.subcommands[cmd.name] = cmd

        for alias in cmd.aliases:
            if self.subcommands.get(alias):
                raise ValueError(
                    f"Duplicate command! Multiple commands share the name/alias: {self.qualified_name} {cmd.name}."
                )
            self.subcommands[alias] = cmd

    def remove_command(self, name: str) -> None:
        """
        Removes a command as a subcommand from this command.

        If an alias is specified, only the alias will be removed.

        Args:
            name: The command to remove.
        """
        command = self.subcommands.pop(name, None)

        if command is None:
            return

        if name in command.aliases:
            command.aliases.remove(name)
            return

        for alias in command.aliases:
            self.subcommands.pop(alias, None)

    def get_command(self, name: str) -> Optional[Self]:
        """
        Gets a subcommand from this command. Can get subcommands of subcommands if needed.

        Args:
            name: The command to search for.

        Returns:
            PrefixedCommand: The command object, if found.
        """
        if " " not in name:
            return self.subcommands.get(name)

        names = name.split()
        if not names:
            return None

        cmd = self.subcommands.get(names[0])
        if not cmd or not cmd.subcommands:
            return cmd

        for name in names[1:]:
            try:
                cmd = cmd.subcommands[name]
            except (AttributeError, KeyError):
                return None

        return cmd

    def subcommand(
        self,
        name: Optional[str] = None,
        *,
        aliases: Optional[list[str]] = None,
        help: Optional[str] = None,
        brief: Optional[str] = None,
        usage: Optional[str] = None,
        enabled: bool = True,
        hidden: bool = False,
        ignore_extra: bool = True,
        inherit_checks: bool = True,
    ) -> Callable[..., Self]:
        """
        A decorator to declare a subcommand for a prefixed command.

        Args:
            name: The name of the command. Defaults to the name of the coroutine.
            aliases: The list of aliases the command can be invoked under.
            help: The long help text for the command. Defaults to the docstring of the coroutine, if there is one.
            brief: The short help text for the command. Defaults to the first line of the help text, if there is one.
            usage: A string displaying how the command can be used. If no string is set, it will \
                default to the command's signature. Useful for help commands.
            enabled: Whether this command can be run at all.
            hidden: If `True`, the default help command (when it is added) does not show this in the help output.
            ignore_extra: If `True`, ignores extraneous strings passed to a command if all its requirements are met \
                (e.g. ?foo a b c when only expecting a and b). Otherwise, an error is raised.
            inherit_checks: If `True`, the subcommand will inherit its checks from the parent command.
        """

        def wrapper(func: Callable) -> Self:
            cmd = PrefixedCommand(
                callback=func,
                name=name or func.__name__,
                aliases=aliases or [],
                help=help,
                brief=brief,
                parent=self,
                usage=usage,
                enabled=enabled,
                hidden=hidden,
                ignore_extra=ignore_extra,
                checks=self.checks if inherit_checks else [],
            )
            self.add_command(cmd)
            return cmd

        return wrapper

    async def call_callback(self, callback: Callable, ctx: "PrefixedContext") -> None:  # noqa: C901
        """
        Runs the callback of this command.

        Args:
            callback (Callable: The callback to run. This is provided for compatibility with internal.
            ctx (internal.MessageContext): The context to use for this command.
        """
        # sourcery skip: remove-empty-nested-block, remove-redundant-if, remove-unnecessary-else
        if len(self.parameters) == 0:
            if ctx.args and not self.ignore_extra:
                raise BadArgument(f"Too many arguments passed to {self.name}.")
            return await self.call_with_binding(callback, ctx)
        # this is slightly costly, but probably worth it
        new_args: list[Any] = []
        kwargs: dict[str, Any] = {}
        args = _PrefixedArgsIterator(tuple(ctx.args))
        param_index = 0

        for arg in args:
            while param_index < len(self.parameters):
                param = self.parameters[param_index]

                if param.consume_rest:
                    arg = args.consume_rest()

                if param.variable:
                    args_to_convert = args.get_rest_of_args()
                    new_arg = [await _convert(param, ctx, arg) for arg in args_to_convert]
                    new_arg = tuple(arg[0] for arg in new_arg)

                    if param.kind == inspect.Parameter.VAR_POSITIONAL:
                        new_args.extend(new_arg)
                    elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                        new_args.append(new_arg)
                    else:
                        kwargs[param.name] = new_arg

                    param_index += 1
                    break

                if param.greedy:
                    greedy_args, broke_off = await _greedy_convert(param, ctx, args)

                    new_args.append(greedy_args)
                    param_index += 1
                    if broke_off:
                        args.back()

                    if param.default:
                        continue
                    break

                converted, used_default = await _convert(param, ctx, arg)
                if param.kind in {
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.VAR_POSITIONAL,
                }:
                    new_args.append(converted)
                else:
                    kwargs[param.name] = converted
                param_index += 1

                if not used_default:
                    break

        if param_index < len(self.parameters):
            for param in self.parameters[param_index:]:
                if param.no_argument:
                    converted, _ = await _convert(param, ctx, None)  # type: ignore
                    if not param.consume_rest:
                        new_args.append(converted)
                    else:
                        kwargs[param.name] = converted
                        break
                    continue

                if not param.optional:
                    raise BadArgument(f"{param.name} is a required argument that is missing.")
                if param.kind in {
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.VAR_POSITIONAL,
                }:
                    new_args.append(param.default)
                else:
                    kwargs[param.name] = param.default
                    break
        elif not self.ignore_extra and not args.finished:
            raise BadArgument(f"Too many arguments passed to {self.name}.")

        return await self.call_with_binding(callback, ctx, *new_args, **kwargs)


def prefixed_command(
    name: Optional[str] = None,
    *,
    aliases: Optional[list[str]] = None,
    help: Optional[str] = None,
    brief: Optional[str] = None,
    usage: Optional[str] = None,
    enabled: bool = True,
    hidden: bool = False,
    ignore_extra: bool = True,
) -> Callable[..., PrefixedCommand]:
    """
    A decorator to declare a coroutine as a prefixed command.

    Args:
        name: The name of the command. Defaults to the name of the coroutine.
        aliases: The list of aliases the command can be invoked under.
        help: The long help text for the command. Defaults to the docstring of the coroutine, if there is one.
        brief: The short help text for the command. Defaults to the first line of the help text, if there is one.
        usage: A string displaying how the command can be used. If no string is set, it will default to the \
            command's signature. Useful for help commands.
        enabled: Whether this command can be run at all.
        hidden: If `True`, the default help command (when it is added) does not show this in the help output.
        ignore_extra: If `True`, ignores extraneous strings passed to a command if all its requirements are \
            met (e.g. ?foo a b c when only expecting a and b). Otherwise, an error is raised.
    """

    def wrapper(func: Callable) -> PrefixedCommand:
        return PrefixedCommand(
            callback=func,
            name=name or func.__name__,
            aliases=aliases or [],
            help=help,
            brief=brief,
            usage=usage,
            enabled=enabled,
            hidden=hidden,
            ignore_extra=ignore_extra,
        )

    return wrapper



from typing import Awaitable, Callable

from interactions.models.discord.role import Role
from interactions.models.discord.snowflake import Snowflake_Type, to_snowflake
from interactions.models.discord.user import Member
from interactions.models.internal.context import BaseContext

__all__ = ("has_role", "has_any_role", "has_id", "is_owner", "guild_only", "dm_only")

TYPE_CHECK_FUNCTION = Callable[[BaseContext], Awaitable[bool]]


def has_role(role: Snowflake_Type | Role) -> TYPE_CHECK_FUNCTION:
    """
    Check if the user has the given role.

    Args:
        role: The Role or role id to check for

    """

    async def check(ctx: BaseContext) -> bool:
        if ctx.guild is None:
            return False
        author: Member = ctx.author  # pyright: ignore [reportGeneralTypeIssues]
        return author.has_role(role)

    return check


def has_any_role(*roles: Snowflake_Type | Role) -> TYPE_CHECK_FUNCTION:
    """
    Checks if the user has any of the given roles.

    Args:
        *roles: The Role(s) or role id(s) to check for
    """

    async def check(ctx: BaseContext) -> bool:
        if ctx.guild is None:
            return False

        author: Member = ctx.author  # pyright: ignore [reportGeneralTypeIssues]
        return any((author.has_role(to_snowflake(r)) for r in roles))

    return check


def has_id(user_id: int) -> TYPE_CHECK_FUNCTION:
    """
    Checks if the author has the desired ID.

    Args:
        user_id: id of the user to check for

    """

    async def check(ctx: BaseContext) -> bool:
        return ctx.author.id == user_id

    return check


def is_owner() -> TYPE_CHECK_FUNCTION:
    """Checks if the author is the owner of the bot. This respects the `client.owner_ids` list."""

    async def check(ctx: BaseContext) -> bool:
        _owner_ids: set = ctx.bot.owner_ids.copy()
        if ctx.bot.app.team:
            [_owner_ids.add(m.id) for m in ctx.bot.app.team.members]
        return ctx.author.id in _owner_ids

    return check


def guild_only() -> TYPE_CHECK_FUNCTION:
    """This command may only be ran in a guild."""

    async def check(ctx: BaseContext) -> bool:
        return ctx.guild is not None

    return check


def dm_only() -> TYPE_CHECK_FUNCTION:
    """This command may only be ran in a dm."""

    async def check(ctx: BaseContext) -> bool:
        return ctx.guild is None

    return check



"""Outlines the interaction between interactions and Discord's Gateway API."""
import asyncio
import sys
import time
import zlib
from asyncio import Task
from types import TracebackType
from typing import TypeVar, TYPE_CHECKING

from interactions.api import events
from interactions.client.const import MISSING, __api_version__
from interactions.client.utils.input_utils import FastJson
from interactions.client.utils.serializer import dict_filter_none
from interactions.models.discord.enums import Status
from interactions.models.discord.enums import WebSocketOPCode as OPCODE
from interactions.models.discord.snowflake import to_snowflake
from interactions.models.internal.cooldowns import CooldownSystem
from .websocket import WebsocketClient

if TYPE_CHECKING:
    from .state import ConnectionState
    from interactions.models.discord.snowflake import Snowflake_Type

__all__ = ("GatewayClient",)


SELF = TypeVar("SELF", bound="WebsocketClient")


class GatewayRateLimit:
    def __init__(self) -> None:
        self.lock = asyncio.Lock()
        # docs state 120 calls per 60 seconds, this is set conservatively to 110 per 60 seconds.
        rate = 110
        interval = 60
        self.cooldown_system = CooldownSystem(1, interval / rate)
        # hacky way to throttle how frequently we send messages to the gateway

    async def rate_limit(self) -> None:
        async with self.lock:
            while not self.cooldown_system.acquire_token():
                await asyncio.sleep(self.cooldown_system.get_cooldown_time())


class GatewayClient(WebsocketClient):
    """
    Abstraction over one gateway connection.

    Multiple `WebsocketClient` instances can be used to implement same-process sharding.

    Attributes:
        sequence: The sequence of this connection
        session_id: The session ID of this connection

    """

    def __init__(self, state: "ConnectionState", shard: tuple[int, int]) -> None:
        super().__init__(state)

        self.shard = shard

        self.chunk_cache = {}

        self._trace = []
        self.sequence = None
        self.session_id = None

        self.ws_url = state.gateway_url
        self.ws_resume_url = MISSING

        # This lock needs to be held to send something over the gateway, but is also held when
        # reconnecting. That way there's no race conditions between sending and reconnecting.
        self._race_lock = asyncio.Lock()
        # Then this event is used so that receive() can wait for the reconnecting to complete.
        self._closed = asyncio.Event()

        self._keep_alive = None
        self._kill_bee_gees = asyncio.Event()
        self._last_heartbeat = 0
        self._acknowledged = asyncio.Event()
        self._acknowledged.set()  # Initialize it as set

        self._ready = asyncio.Event()
        self._close_gateway = asyncio.Event()

        # Sanity check, it is extremely important that an instance isn't reused.
        self._entered = False

    async def __aenter__(self: SELF) -> SELF:
        if self._entered:
            raise RuntimeError("An instance of 'WebsocketClient' cannot be re-used!")

        self._entered = True
        self._zlib = zlib.decompressobj()

        self.ws = await self.state.client.http.websocket_connect(self.state.gateway_url)

        hello = await self.receive(force=True)
        self.heartbeat_interval = hello["d"]["heartbeat_interval"] / 1000
        self._closed.set()

        self._keep_alive = asyncio.create_task(self.run_bee_gees())

        await self._identify()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        # Technically should not be possible in any way, but might as well be safe worst-case.
        self._close_gateway.set()

        try:
            if self._keep_alive is not None:
                self._kill_bee_gees.set()
                try:
                    # Even if we get cancelled that is fine, because then the keep-alive
                    # handler will also be cancelled since we're waiting on it.
                    await self._keep_alive  # Wait for the keep-alive handler to finish
                finally:
                    self._keep_alive = None
        finally:
            if self.ws is not None:
                # We could be cancelled here, it is extremely important that we close the
                # WebSocket either way, hence the try/except.
                try:
                    await self.ws.close(code=1000)
                finally:
                    self.ws = None

    async def run(self) -> None:
        """Start receiving events from the websocket."""
        while True:
            if self._stopping is None:
                self._stopping = asyncio.create_task(self._close_gateway.wait())
            receiving = asyncio.create_task(self.receive())
            done, _ = await asyncio.wait({self._stopping, receiving}, return_when=asyncio.FIRST_COMPLETED)

            if receiving in done:
                # Note that we check for a received message first, because if both completed at
                # the same time, we don't want to discard that message.
                msg = await receiving
            else:
                # This has to be the stopping task, which we join into the current task (even
                # though that doesn't give any meaningful value in the return).
                await self._stopping
                receiving.cancel()
                return

            op = msg.get("op")
            data = msg.get("d")
            seq = msg.get("s")
            event = msg.get("t")

            if seq:
                self.sequence = seq

            if op == OPCODE.DISPATCH:
                _ = asyncio.create_task(self.dispatch_event(data, seq, event))
                continue

            # This may try to reconnect the connection so it is best to wait
            # for it to complete before receiving more - that way there's less
            # possible race conditions to consider.
            await self.dispatch_opcode(data, op)

    async def dispatch_opcode(self, data, op: OPCODE) -> None:
        match op:
            case OPCODE.HEARTBEAT:
                self.logger.debug("Received heartbeat request from gateway")
                return await self.send_heartbeat()

            case OPCODE.HEARTBEAT_ACK:
                self._latency.append(time.perf_counter() - self._last_heartbeat)

                if self._last_heartbeat != 0 and self._latency[-1] >= 15:
                    self.logger.warning(
                        f"High Latency! shard ID {self.shard[0]} heartbeat took {self._latency[-1]:.1f}s to be acknowledged!"
                    )
                else:
                    self.logger.debug(f"❤ Heartbeat acknowledged after {self._latency[-1]:.5f} seconds")

                return self._acknowledged.set()

            case OPCODE.RECONNECT:
                self.logger.debug("Gateway requested reconnect. Reconnecting...")
                return await self.reconnect(resume=True, url=self.ws_resume_url)

            case OPCODE.INVALIDATE_SESSION:
                self.logger.warning("Gateway has invalidated session! Reconnecting...")
                return await self.reconnect()

            case _:
                return self.logger.debug(f"Unhandled OPCODE: {op} = {OPCODE(op).name}")

    async def dispatch_event(self, data, seq, event) -> None:
        match event:
            case "READY":
                self._ready.set()
                self._trace = data.get("_trace", [])
                self.sequence = seq
                self.session_id = data["session_id"]
                self.ws_resume_url = (
                    f"{data['resume_gateway_url']}?encoding=json&v={__api_version__}&compress=zlib-stream"
                )
                self.logger.info(f"Shard {self.shard[0]} has connected to gateway!")
                self.logger.debug(f"Session ID: {self.session_id} Trace: {self._trace}")
                return self.state.client.dispatch(events.WebsocketReady(data))

            case "RESUMED":
                self.logger.info(f"Successfully resumed connection! Session_ID: {self.session_id}")
                self.state.client.dispatch(events.Resume())
                return None

            case "GUILD_MEMBERS_CHUNK":
                _ = asyncio.create_task(self._process_member_chunk(data.copy()))

            case _:
                # the above events are "special", and are handled by the gateway itself, the rest can be dispatched
                event_name = f"raw_{event.lower()}"
                if processor := self.state.client.processors.get(event_name):
                    try:
                        _ = asyncio.create_task(
                            processor(events.RawGatewayEvent(data.copy(), override_name=event_name))
                        )
                    except Exception as ex:
                        self.logger.error(f"Failed to run event processor for {event_name}: {ex}")
                else:
                    self.logger.debug(f"No processor for `{event_name}`")

        self.state.client.dispatch(events.RawGatewayEvent(data.copy(), override_name="raw_gateway_event"))
        self.state.client.dispatch(events.RawGatewayEvent(data.copy(), override_name=f"raw_{event.lower()}"))

    def close(self) -> None:
        """Shutdown the websocket connection."""
        self._close_gateway.set()

    async def _identify(self) -> None:
        """Send an identify payload to the gateway."""
        if self.ws is None:
            raise RuntimeError
        payload = {
            "op": OPCODE.IDENTIFY,
            "d": {
                "token": self.state.client.http.token,
                "intents": self.state.intents,
                "shard": self.shard,
                "large_threshold": 250,
                "properties": {
                    "os": sys.platform,
                    "browser": "interactions",
                    "device": "interactions",
                },
                "presence": self.state.presence,
            },
            "compress": True,
        }

        serialized = FastJson.dumps(payload)
        await self.ws.send_str(serialized)

        self.logger.debug(
            f"Shard ID {self.shard[0]} has identified itself to Gateway, requesting intents: {self.state.intents}!"
        )

    async def reconnect(self, *, resume: bool = False, code: int = 1012, url: str | None = None) -> None:
        self.state.clear_ready()
        self._ready.clear()
        await super().reconnect(resume=resume, code=code, url=url)

    async def _resume_connection(self) -> None:
        """Send a resume payload to the gateway."""
        if self.ws is None:
            raise RuntimeError

        payload = {
            "op": OPCODE.RESUME,
            "d": {
                "token": self.state.client.http.token,
                "seq": self.sequence,
                "session_id": self.session_id,
            },
        }

        serialized = FastJson.dumps(payload)
        await self.ws.send_str(serialized)

        self.logger.debug(f"{self.shard[0]} is attempting to resume a connection")

    async def send_heartbeat(self) -> None:
        await self.send_json({"op": OPCODE.HEARTBEAT, "d": self.sequence}, bypass=True)
        self.logger.debug(f"❤ Shard {self.shard[0]} is sending a Heartbeat")

    async def change_presence(self, activity=None, status: Status = Status.ONLINE, since=None) -> None:
        """Update the bot's presence status."""
        await self.send_json(
            {
                "op": OPCODE.PRESENCE,
                "d": dict_filter_none(
                    {
                        "since": int(since or time.time() * 1000),
                        "activities": [activity] if activity else [],
                        "status": status,
                        "afk": False,
                    }
                ),
            }
        )

    async def request_member_chunks(
        self,
        guild_id: "Snowflake_Type",
        query="",
        *,
        limit,
        user_ids=None,
        presences=False,
        nonce=None,
    ) -> None:
        payload = {
            "op": OPCODE.REQUEST_MEMBERS,
            "d": dict_filter_none(
                {
                    "guild_id": guild_id,
                    "presences": presences,
                    "limit": limit,
                    "nonce": nonce,
                    "user_ids": user_ids,
                    "query": query,
                }
            ),
        }
        await self.send_json(payload)

    async def _process_member_chunk(self, chunk: dict) -> Task[None]:
        if guild := self.state.client.cache.get_guild(to_snowflake(chunk.get("guild_id"))):
            return asyncio.create_task(guild.process_member_chunk(chunk))
        raise ValueError(f"No guild exists for {chunk.get('guild_id')}")

    async def voice_state_update(
        self,
        guild_id: "Snowflake_Type",
        channel_id: "Snowflake_Type",
        muted: bool = False,
        deafened: bool = False,
    ) -> None:
        """Update the bot's voice state."""
        payload = {
            "op": OPCODE.VOICE_STATE,
            "d": {
                "guild_id": guild_id,
                "channel_id": channel_id,
                "self_mute": muted,
                "self_deaf": deafened,
            },
        }
        await self.send_json(payload)



"""
Sets up a Sentry Logger

And then call `bot.load_extension('interactions.ext.sentry', token=SENTRY_TOKEN)`
Optionally takes a filter function that will be called before sending the event to Sentry.
"""
import functools
import logging
from typing import Any, Callable, Optional

from interactions.api.events.internal import Error
from interactions.client.const import get_logger
from interactions.models.internal.tasks.task import Task

try:
    import sentry_sdk
except ModuleNotFoundError:
    get_logger().error(
        "sentry-sdk not installed, cannot enable sentry integration.  Install with `pip install discord-py-interactions[sentry]`"
    )
    raise

from interactions import Client, Extension, listen

__all__ = ("setup", "default_sentry_filter")


def default_sentry_filter(event: dict[str, Any], hint: dict[str, Any]) -> Optional[dict[str, Any]]:
    if "log_record" in hint:
        record: logging.LogRecord = hint["log_record"]
        if "interactions" in record.name:
            #  There are some logging messages that are not worth sending to sentry.
            if ": 403" in record.message:
                return None
            if record.message.startswith("Ignoring exception in "):
                return None

    if "exc_info" in hint:
        exc_type, exc_value, tb = hint["exc_info"]
        if isinstance(exc_value, KeyboardInterrupt):
            #  We don't need to report a ctrl+c
            return None
    return event


class SentryExtension(Extension):
    @listen()
    async def on_startup(self) -> None:
        sentry_sdk.set_context(
            "bot",
            {
                "name": str(self.bot.user),
                "intents": repr(self.bot.intents),
            },
        )
        sentry_sdk.set_tag("bot_name", str(self.bot.user))

    @listen(disable_default_listeners=False)
    async def on_error(self, event: Error) -> None:
        with sentry_sdk.configure_scope() as scope:
            scope.set_tag("source", event.source)
            if event.ctx:
                scope.set_context(
                    type(event.ctx).__name__,
                    {
                        "args": event.ctx.args,
                        "kwargs": event.ctx.kwargs,
                        "message": event.ctx.message,
                    },
                )
                if event.ctx.author:
                    scope.set_user({"id": event.ctx.author.id, "username": event.ctx.author.tag})
            sentry_sdk.capture_exception(event.error)


class HookedTask(Task):
    """We're subclassing purely for the type hinting.  The following method will be transplanted onto Task."""

    def on_error_sentry_hook(self: Task, error: Exception) -> None:
        with sentry_sdk.configure_scope() as scope:
            if isinstance(self.callback, functools.partial):
                scope.set_tag("task", self.callback.func.__name__)
            else:
                scope.set_tag("task", self.callback.__name__)

            scope.set_tag("iteration", self.iteration)
            sentry_sdk.capture_exception(error)


def setup(
    bot: Client,
    token: str | None = None,
    filter: Optional[Callable[[dict[str, Any], dict[str, Any]], Optional[dict[str, Any]]]] = None,
    **kwargs,
) -> None:
    if not token:
        bot.logger.error("Cannot enable sentry integration, no token provided")
        return
    if filter is None:
        filter = default_sentry_filter
    sentry_sdk.init(token, before_send=filter, **kwargs)
    Task.on_error_sentry_hook = HookedTask.on_error_sentry_hook  # type: ignore
    SentryExtension(bot)



import asyncio
import inspect
from typing import Callable

from interactions.api.events.internal import BaseEvent
from interactions.client.const import MISSING, Absent, AsyncCallable
from interactions.client.utils import get_event_name
from interactions.models.internal.callback import CallbackObject

__all__ = ("Listener", "listen")


class Listener(CallbackObject):
    event: str
    """Name of the event to listen to."""
    callback: AsyncCallable
    """Coroutine to call when the event is triggered."""
    is_default_listener: bool
    """Whether this listener is provided automatically by the library, and might be unwanted by users."""
    disable_default_listeners: bool
    """Whether this listener supersedes default listeners.  If true, any default listeners will be unregistered."""
    delay_until_ready: bool
    """whether to delay the event until the client is ready"""

    def __init__(
        self,
        func: AsyncCallable,
        event: str,
        *,
        delay_until_ready: bool = False,
        is_default_listener: bool = False,
        disable_default_listeners: bool = False,
        pass_event_object: Absent[bool] = MISSING,
    ) -> None:
        super().__init__()

        if is_default_listener:
            disable_default_listeners = False

        self.event = event
        self.callback = func
        self.delay_until_ready = delay_until_ready
        self.is_default_listener = is_default_listener
        self.disable_default_listeners = disable_default_listeners

        self._params = inspect.signature(func).parameters.copy()
        self.pass_event_object = pass_event_object
        self.warned_no_event_arg = False

    def __repr__(self) -> str:
        return f"<Listener event={self.event!r} callback={self.callback!r}>"

    @classmethod
    def create(
        cls,
        event_name: Absent[str | BaseEvent] = MISSING,
        *,
        delay_until_ready: bool = False,
        is_default_listener: bool = False,
        disable_default_listeners: bool = False,
    ) -> Callable[[AsyncCallable], "Listener"]:
        """
        Decorator for creating an event listener.

        Args:
            event_name: The name of the event to listen to. If left blank, event name will be inferred from the function name or parameter.
            delay_until_ready: Whether to delay the listener until the client is ready.
            is_default_listener: Whether this listener is provided automatically by the library, and might be unwanted by users.
            disable_default_listeners: Whether this listener supersedes default listeners.  If true, any default listeners will be unregistered.


        Returns:
            A listener object.

        """

        def wrapper(coro: AsyncCallable) -> "Listener":
            if not asyncio.iscoroutinefunction(coro):
                raise TypeError("Listener must be a coroutine")

            name = event_name

            if name is MISSING:
                for typehint in coro.__annotations__.values():
                    if (
                        inspect.isclass(typehint)
                        and issubclass(typehint, BaseEvent)
                        and typehint.__name__ != "RawGatewayEvent"
                    ):
                        name = typehint.__name__
                        break

                if not name:
                    name = coro.__name__

            return cls(
                coro,
                get_event_name(name),
                delay_until_ready=delay_until_ready,
                is_default_listener=is_default_listener,
                disable_default_listeners=disable_default_listeners,
            )

        return wrapper

    def lazy_parse_params(self):
        """Process the parameters of this listener."""
        if self.pass_event_object is not MISSING:
            return

        if self.has_binding:
            # discard the first parameter, which is the class instance
            self._params = list(self._params.values())[1:]

        self.pass_event_object = len(self._params) != 0


def listen(
    event_name: Absent[str | type[BaseEvent]] = MISSING,
    *,
    delay_until_ready: bool = False,
    is_default_listener: bool = False,
    disable_default_listeners: bool = False,
) -> Callable[[AsyncCallable], Listener]:
    """
    Decorator to make a function an event listener.

    Args:
        event_name: The name of the event to listen to. If left blank, event name will be inferred from the function name or parameter.
        delay_until_ready: Whether to delay the listener until the client is ready.
        is_default_listener: Whether this listener is provided automatically by the library, and might be unwanted by users.
        disable_default_listeners: Whether this listener supersedes default listeners.  If true, any default listeners will be unregistered.


    Returns:
        A listener object.

    """
    return Listener.create(
        event_name,
        delay_until_ready=delay_until_ready,
        is_default_listener=is_default_listener,
        disable_default_listeners=disable_default_listeners,
    )



import re
import string
import unicodedata
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import attrs
import emoji

from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.attr_converters import list_converter
from interactions.client.utils.attr_converters import optional
from interactions.client.utils.serializer import dict_filter_none, no_export_meta
from interactions.models.discord.base import ClientObject
from interactions.models.discord.snowflake import SnowflakeObject, to_snowflake, to_snowflake_list

if TYPE_CHECKING:
    from interactions.client import Client
    from interactions.models.discord.guild import Guild
    from interactions.models.discord.user import User, Member
    from interactions.models.discord.role import Role
    from interactions.models.discord.snowflake import Snowflake_Type

__all__ = ("PartialEmoji", "CustomEmoji", "process_emoji_req_format", "process_emoji")

emoji_regex = re.compile(r"<?(a)?:(\w*):(\d*)>?")
unicode_emoji_reg = re.compile(r"[^\w\s,’‘“”…–—•◦‣⁃⁎⁏⁒⁓⁺⁻⁼⁽⁾ⁿ₊₋₌₍₎]")  # noqa: RUF001


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class PartialEmoji(SnowflakeObject, DictSerializationMixin):
    """Represent a basic ("partial") emoji used in discord."""

    id: Optional["Snowflake_Type"] = attrs.field(
        repr=True, default=None, converter=optional(to_snowflake)
    )  # can be None for Standard Emoji
    """The custom emoji id. Leave empty if you are using standard unicode emoji."""
    name: Optional[str] = attrs.field(repr=True, default=None)
    """The custom emoji name, or standard unicode emoji in string"""
    animated: bool = attrs.field(repr=True, default=False)
    """Whether this emoji is animated"""

    @classmethod
    def from_str(cls, emoji_str: str, *, language: str = "alias") -> Optional["PartialEmoji"]:
        """
        Generate a PartialEmoji from a discord Emoji string representation, or unicode emoji.

        Handles:
            <:emoji_name:emoji_id>
            :emoji_name:emoji_id
            <a:emoji_name:emoji_id>
            a:emoji_name:emoji_id
            👋
            :wave:

        Args:
            emoji_str: The string representation an emoji
            language: The language to use for the unicode emoji parsing

        Returns:
            A PartialEmoji object

        Raises:
            ValueError: if the string cannot be parsed

        """
        if parsed := emoji_regex.findall(emoji_str):
            parsed = tuple(filter(None, parsed[0]))
            if len(parsed) == 3:
                return cls(name=parsed[1], id=parsed[2], animated=True)
            if len(parsed) == 2:
                return cls(name=parsed[0], id=parsed[1])
            _name = emoji.emojize(emoji_str, language=language)
            if _emoji_list := emoji.distinct_emoji_list(_name):
                return cls(name=_emoji_list[0])
        else:
            if _emoji_list := emoji.distinct_emoji_list(emoji_str):
                return cls(name=_emoji_list[0])

            # the emoji lib handles *most* emoji, however there are certain ones that it misses
            # this acts as a fallback check
            if matches := unicode_emoji_reg.search(emoji_str):
                match = matches.group()

                # the regex will match certain special characters, so this acts as a final failsafe
                if match not in string.printable and unicodedata.category(match) == "So":
                    return cls(name=match)
        return None

    def __str__(self) -> str:
        s = self.req_format
        if self.id:
            s = f"<{'a:' if self.animated else ':'}{s}>"
        return s

    def __eq__(self, other) -> bool:
        if not isinstance(other, PartialEmoji):
            return False
        return self.id == other.id if self.id else self.name == other.name

    @property
    def req_format(self) -> str:
        """Format used for web request."""
        return f"{self.name}:{self.id}" if self.id else self.name


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class CustomEmoji(PartialEmoji, ClientObject):
    """Represent a custom emoji in a guild with all its properties."""

    _client: "Client" = attrs.field(repr=False, metadata=no_export_meta)

    require_colons: bool = attrs.field(repr=False, default=False)
    """Whether this emoji must be wrapped in colons"""
    managed: bool = attrs.field(repr=False, default=False)
    """Whether this emoji is managed"""
    available: bool = attrs.field(repr=False, default=False)
    """Whether this emoji can be used, may be false due to loss of Server Boosts."""

    _creator_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=optional(to_snowflake))
    _role_ids: List["Snowflake_Type"] = attrs.field(
        repr=False, factory=list, converter=optional(list_converter(to_snowflake))
    )
    _guild_id: "Snowflake_Type" = attrs.field(repr=False, default=None, converter=to_snowflake)

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Client") -> Dict[str, Any]:
        creator_dict = data.pop("user", None)
        data["creator_id"] = client.cache.place_user_data(creator_dict).id if creator_dict else None

        if "roles" in data:
            data["role_ids"] = data.pop("roles")

        return data

    @classmethod
    def from_dict(cls, data: Dict[str, Any], client: "Client", guild_id: int) -> "CustomEmoji":
        data = cls._process_dict(data, client)
        return cls(client=client, guild_id=guild_id, **cls._filter_kwargs(data, cls._get_init_keys()))

    @property
    def guild(self) -> "Guild":
        """The guild this emoji belongs to."""
        return self._client.cache.get_guild(self._guild_id)

    @property
    def creator(self) -> Optional[Union["Member", "User"]]:
        """The member that created this emoji."""
        return self._client.cache.get_member(self._creator_id, self._guild_id) or self._client.cache.get_user(
            self._creator_id
        )

    @property
    def roles(self) -> List["Role"]:
        """The roles allowed to use this emoji."""
        return [self._client.cache.get_role(role_id) for role_id in self._role_ids]

    @property
    def is_usable(self) -> bool:
        """Determines if this emoji is usable by the current user."""
        if not self.available:
            return False

        guild = self.guild
        return any(e_role_id in guild.me._role_ids for e_role_id in self._role_ids)

    async def edit(
        self,
        *,
        name: Optional[str] = None,
        roles: Optional[List[Union["Snowflake_Type", "Role"]]] = None,
        reason: Optional[str] = None,
    ) -> "CustomEmoji":
        """
        Modify the custom emoji information.

        Args:
            name: The name of the emoji.
            roles: The roles allowed to use this emoji.
            reason: Attach a reason to this action, used for audit logs.

        Returns:
            The newly modified custom emoji.

        """
        data_payload = dict_filter_none(
            {
                "name": name,
                "roles": to_snowflake_list(roles) if roles else None,
            }
        )

        updated_data = await self._client.http.modify_guild_emoji(data_payload, self._guild_id, self.id, reason=reason)
        self.update_from_dict(updated_data)
        return self

    async def delete(self, reason: Optional[str] = None) -> None:
        """
        Deletes the custom emoji from the guild.

        Args:
            reason: Attach a reason to this action, used for audit logs.

        """
        if not self._guild_id:
            raise ValueError("Cannot delete emoji, no guild id set.")

        await self._client.http.delete_guild_emoji(self._guild_id, self.id, reason=reason)

    @property
    def url(self) -> str:
        """CDN url for the emoji."""
        return f"https://cdn.discordapp.net/emojis/{self.id}.{'gif' if self.animated else 'png'}"


def process_emoji_req_format(emoji: Optional[Union[PartialEmoji, dict, str]]) -> Optional[str]:
    """
    Processes the emoji parameter into the str format required by the API.

    Args:
        emoji: The emoji to process.

    Returns:
        formatted string for discord

    """
    if not emoji:
        return emoji

    if isinstance(emoji, str):
        emoji = PartialEmoji.from_str(emoji)

    if isinstance(emoji, dict):
        emoji = PartialEmoji.from_dict(emoji)

    if isinstance(emoji, PartialEmoji):
        return emoji.req_format

    raise ValueError(f"Invalid emoji: {emoji}")


def process_emoji(emoji: Optional[Union[PartialEmoji, dict, str]]) -> Optional[dict]:
    """
    Processes the emoji parameter into the dictionary format required by the API.

    Args:
        emoji: The emoji to process.

    Returns:
        formatted dictionary for discord

    """
    if not emoji:
        return emoji

    if isinstance(emoji, dict):
        return emoji

    if isinstance(emoji, str):
        emoji = PartialEmoji.from_str(emoji)

    if isinstance(emoji, PartialEmoji):
        return emoji.to_dict()

    raise ValueError(f"Invalid emoji: {emoji}")



import contextlib
import uuid
from abc import abstractmethod
from typing import Any, Dict, Iterator, List, Optional, Union

import discord_typings

from interactions.client.const import ACTION_ROW_MAX_ITEMS, MISSING
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.models.discord.emoji import PartialEmoji, process_emoji
from interactions.models.discord.enums import ButtonStyle, ChannelType, ComponentType

__all__ = (
    "BaseComponent",
    "InteractiveComponent",
    "ActionRow",
    "Button",
    "BaseSelectMenu",
    "StringSelectMenu",
    "StringSelectOption",
    "UserSelectMenu",
    "RoleSelectMenu",
    "MentionableSelectMenu",
    "ChannelSelectMenu",
    "process_components",
    "spread_to_rows",
    "get_components_ids",
    "TYPE_COMPONENT_MAPPING",
)


class BaseComponent(DictSerializationMixin):
    """
    A base component class.

    !!! Warning
        This should never be directly instantiated.

    """

    type: ComponentType

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type}>"

    @classmethod
    @abstractmethod
    def from_dict(cls, data: dict) -> "BaseComponent":
        """
        Create a component from a dictionary.

        Args:
            data: the dictionary to create the component from

        Returns:
            The created component.

        """
        raise NotImplementedError

    @classmethod
    def from_dict_factory(
        cls,
        data: dict,
        *,
        alternate_mapping: dict[ComponentType, "BaseComponent"] | None = None,
    ) -> "BaseComponent":
        """
        Creates a component from a payload.

        Args:
            data: the payload from Discord
            alternate_mapping: an optional mapping of component types to classes
        """
        data.pop("hash", None)  # redundant

        component_type = data.pop("type", None)

        mapping = alternate_mapping or TYPE_COMPONENT_MAPPING

        if component_class := mapping.get(component_type, None):
            return component_class.from_dict(data)
        raise TypeError(f"Unsupported component type for {data} ({component_type}), please consult the docs.")


class InteractiveComponent(BaseComponent):
    """
    A base interactive component class.

    !!! Warning
        This should never be instantiated.

    """

    type: ComponentType
    custom_id: str

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, dict):
            other = BaseComponent.from_dict_factory(other)
        return self.custom_id == other.custom_id and self.type == other.type

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type} custom_id={self.custom_id}>"


class ActionRow(BaseComponent):
    """
    Represents an action row.

    Attributes:
        components list[Dict | BaseComponent]: A sequence of components contained within this action row
    """

    def __init__(self, *components: Dict | BaseComponent) -> None:
        if isinstance(components, (list, tuple)):
            # flatten user error
            components = list(components)

        self.components: list[Dict | BaseComponent] = [
            BaseComponent.from_dict_factory(c) if isinstance(c, dict) else c for c in components
        ]

        self.type: ComponentType = ComponentType.ACTION_ROW
        self._max_items = ACTION_ROW_MAX_ITEMS

    @classmethod
    def from_dict(cls, data: discord_typings.ActionRowData) -> "ActionRow":
        return cls(*data["components"])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type} components={len(self.components)}>"

    def add_component(self, *components: dict | BaseComponent) -> None:
        """
        Add a component to this action row.

        Args:
            *components: The component(s) to add.
        """
        if isinstance(components, (list, tuple)):
            # flatten user error
            components = list(components)

        self.components.extend(BaseComponent.from_dict_factory(c) if isinstance(c, dict) else c for c in components)

    @classmethod
    def split_components(cls, *components: dict | BaseComponent, count_per_row: int = 5) -> list["ActionRow"]:
        """
        Split components into action rows.

        Args:
            *components: The components to split.
            count_per_row: The amount of components to have per row.

        Returns:
            A list of action rows.
        """
        buffer = []
        action_rows = []

        for component in components:
            c_type = component.type if hasattr(component, "type") else component["type"]
            if c_type in (
                ComponentType.STRING_SELECT,
                ComponentType.USER_SELECT,
                ComponentType.ROLE_SELECT,
                ComponentType.MENTIONABLE_SELECT,
                ComponentType.CHANNEL_SELECT,
            ):
                # Selects can only be in their own row
                if buffer:
                    action_rows.append(cls(*buffer))
                    buffer = []
                action_rows.append(cls(component))
            else:
                buffer.append(component)
                if len(buffer) >= count_per_row:
                    action_rows.append(cls(*buffer))
                    buffer = []

        if buffer:
            action_rows.append(cls(*buffer))
        return action_rows

    def to_dict(self) -> discord_typings.ActionRowData:
        return {
            "type": self.type.value,  # type: ignore
            "components": [c.to_dict() for c in self.components],
        }


class Button(InteractiveComponent):
    """
    Represents a discord ui button.

    Attributes:
        style optional[ButtonStyle, int]: Buttons come in a variety of styles to convey different types of actions.
        label optional[str]: The text that appears on the button, max 80 characters.
        emoji optional[Union[PartialEmoji, dict, str]]: The emoji that appears on the button.
        custom_id Optional[str]: A developer-defined identifier for the button, max 100 characters.
        url Optional[str]: A url for link-style buttons.
        disabled bool: Disable the button and make it not interactable, default false.

    """

    Styles: ButtonStyle = ButtonStyle

    def __init__(
        self,
        *,
        style: ButtonStyle | int,
        label: str | None = None,
        emoji: "PartialEmoji | None | str" = None,
        custom_id: str | None = None,
        url: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.style: ButtonStyle = ButtonStyle(style)
        self.label: str | None = label
        self.emoji: "PartialEmoji | None" = emoji
        self.custom_id: str | None = custom_id
        self.url: str | None = url
        self.disabled: bool = disabled

        self.type: ComponentType = ComponentType.BUTTON

        if self.style == ButtonStyle.URL:
            if self.custom_id is not None:
                raise ValueError("URL buttons cannot have a custom_id.")
            if self.url is None:
                raise ValueError("URL buttons must have a url.")

        elif self.custom_id is None:
            self.custom_id = str(uuid.uuid4())
        if not self.label and not self.emoji:
            raise ValueError("Buttons must have a label or an emoji.")

        if isinstance(self.emoji, str):
            self.emoji = PartialEmoji.from_str(self.emoji)

    @classmethod
    def from_dict(cls, data: discord_typings.ButtonComponentData) -> "Button":
        emoji = process_emoji(data.get("emoji"))
        emoji = PartialEmoji.from_dict(emoji) if emoji else None
        return cls(
            style=ButtonStyle(data["style"]),
            label=data.get("label"),
            emoji=emoji,
            custom_id=data.get("custom_id"),
            url=data.get("url"),
            disabled=data.get("disabled", False),
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type} style={self.style} label={self.label} emoji={self.emoji} custom_id={self.custom_id} url={self.url} disabled={self.disabled}>"

    def to_dict(self) -> discord_typings.ButtonComponentData:
        emoji = self.emoji.to_dict() if self.emoji else None
        if emoji and hasattr(emoji, "to_dict"):
            emoji = emoji.to_dict()

        return {
            "type": self.type.value,  # type: ignore
            "style": self.style.value,  # type: ignore
            "label": self.label,
            "emoji": emoji,
            "custom_id": self.custom_id,
            "url": self.url,
            "disabled": self.disabled,
        }


class BaseSelectMenu(InteractiveComponent):
    """
    Represents a select menu component

    Attributes:
        custom_id str: A developer-defined identifier for the button, max 100 characters.
        placeholder str: The custom placeholder text to show if nothing is selected, max 100 characters.
        min_values Optional[int]: The minimum number of items that must be chosen. (default 1, min 0, max 25)
        max_values Optional[int]: The maximum number of items that can be chosen. (default 1, max 25)
        disabled bool: Disable the select and make it not intractable, default false.
        type Union[ComponentType, int]: The action role type number defined by discord. This cannot be modified.
    """

    def __init__(
        self,
        *,
        placeholder: str | None = None,
        min_values: int = 1,
        max_values: int = 1,
        custom_id: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.custom_id: str = custom_id or str(uuid.uuid4())
        self.placeholder: str | None = placeholder
        self.min_values: int = min_values
        self.max_values: int = max_values
        self.disabled: bool = disabled

        self.type: ComponentType = MISSING

    @classmethod
    def from_dict(cls, data: discord_typings.SelectMenuComponentData) -> "BaseSelectMenu":
        return cls(
            placeholder=data.get("placeholder"),
            min_values=data["min_values"],
            max_values=data["max_values"],
            custom_id=data["custom_id"],
            disabled=data.get("disabled", False),
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type} custom_id={self.custom_id} placeholder={self.placeholder} min_values={self.min_values} max_values={self.max_values} disabled={self.disabled}>"

    def to_dict(self) -> discord_typings.SelectMenuComponentData:
        return {
            "type": self.type.value,  # type: ignore
            "custom_id": self.custom_id,
            "placeholder": self.placeholder,
            "min_values": self.min_values,
            "max_values": self.max_values,
            "disabled": self.disabled,
        }


class StringSelectOption(BaseComponent):
    """
    Represents a select option.

    Attributes:
        label str: The label (max 80 characters)
        value str: The value of the select, this is whats sent to your bot
        description Optional[str]: A description of this option
        emoji Optional[Union[PartialEmoji, dict, str]: An emoji to show in this select option
        default bool: Is this option selected by default
    """

    def __init__(
        self,
        *,
        label: str,
        value: str,
        description: str | None = None,
        emoji: "PartialEmoji | None | str" = None,
        default: bool = False,
    ) -> None:
        self.label: str = label
        self.value: str = value
        self.description: str | None = description
        self.emoji: PartialEmoji | None = emoji
        self.default: bool = default

        if isinstance(self.emoji, str):
            self.emoji = PartialEmoji.from_str(self.emoji)

    @classmethod
    def converter(cls, value: Any) -> "StringSelectOption":
        if isinstance(value, StringSelectOption):
            return value
        if isinstance(value, dict):
            return cls.from_dict(value)

        if isinstance(value, str):
            return cls(label=value, value=value)

        with contextlib.suppress(TypeError):
            possible_iter = iter(value)

            return cls(label=possible_iter[0], value=possible_iter[1])
        raise TypeError(f"Cannot convert {value} of type {type(value)} to a SelectOption")

    @classmethod
    def from_dict(cls, data: discord_typings.SelectMenuOptionData) -> "StringSelectOption":
        emoji = process_emoji(data.get("emoji"))
        emoji = PartialEmoji.from_dict(emoji) if emoji else None
        return cls(
            label=data["label"],
            value=data["value"],
            description=data.get("description"),
            emoji=emoji,
            default=data.get("default", False),
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} label={self.label} value={self.value} description={self.description} emoji={self.emoji} default={self.default}>"

    def to_dict(self) -> discord_typings.SelectMenuOptionData:
        emoji = self.emoji.to_dict() if self.emoji else None
        if emoji and hasattr(emoji, "to_dict"):
            emoji = emoji.to_dict()

        return {
            "label": self.label,
            "value": self.value,
            "description": self.description,
            "emoji": emoji,
            "default": self.default,
        }


class StringSelectMenu(BaseSelectMenu):
    """
    Represents a string select component.

    Attributes:
        options List[dict]: The choices in the select, max 25.
        custom_id str: A developer-defined identifier for the button, max 100 characters.
        placeholder str: The custom placeholder text to show if nothing is selected, max 100 characters.
        min_values Optional[int]: The minimum number of items that must be chosen. (default 1, min 0, max 25)
        max_values Optional[int]: The maximum number of items that can be chosen. (default 1, max 25)
        disabled bool: Disable the select and make it not intractable, default false.
        type Union[ComponentType, int]: The action role type number defined by discord. This cannot be modified.
    """

    def __init__(
        self,
        *options: StringSelectOption
        | str
        | discord_typings.SelectMenuOptionData
        | list[StringSelectOption | str | discord_typings.SelectMenuOptionData],
        placeholder: str | None = None,
        min_values: int = 1,
        max_values: int = 1,
        custom_id: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            custom_id=custom_id,
            disabled=disabled,
        )
        if isinstance(options, (list, tuple)) and len(options) == 1 and isinstance(options[0], (list, tuple)):
            # user passed in a list of options, expand it out
            options = options[0]

        self.options: list[StringSelectOption] = [StringSelectOption.converter(option) for option in options]
        self.type: ComponentType = ComponentType.STRING_SELECT

    @classmethod
    def from_dict(cls, data: discord_typings.SelectMenuComponentData) -> "StringSelectMenu":
        return cls(
            *data["options"],
            placeholder=data.get("placeholder"),
            min_values=data["min_values"],
            max_values=data["max_values"],
            custom_id=data["custom_id"],
            disabled=data.get("disabled", False),
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type} custom_id={self.custom_id} placeholder={self.placeholder} min_values={self.min_values} max_values={self.max_values} disabled={self.disabled} options={self.options}>"

    def to_dict(self) -> discord_typings.SelectMenuComponentData:
        return {
            **super().to_dict(),
            "options": [option.to_dict() for option in self.options],
        }


class UserSelectMenu(BaseSelectMenu):
    """
    Represents a user select component.

    Attributes:
        custom_id str: A developer-defined identifier for the button, max 100 characters.
        placeholder str: The custom placeholder text to show if nothing is selected, max 100 characters.
        min_values Optional[int]: The minimum number of items that must be chosen. (default 1, min 0, max 25)
        max_values Optional[int]: The maximum number of items that can be chosen. (default 1, max 25)
        disabled bool: Disable the select and make it not intractable, default false.
        type Union[ComponentType, int]: The action role type number defined by discord. This cannot be modified.
    """

    def __init__(
        self,
        *,
        placeholder: str | None = None,
        min_values: int = 1,
        max_values: int = 1,
        custom_id: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            custom_id=custom_id,
            disabled=disabled,
        )

        self.type: ComponentType = ComponentType.USER_SELECT


class RoleSelectMenu(BaseSelectMenu):
    """
    Represents a user select component.

    Attributes:
        custom_id str: A developer-defined identifier for the button, max 100 characters.
        placeholder str: The custom placeholder text to show if nothing is selected, max 100 characters.
        min_values Optional[int]: The minimum number of items that must be chosen. (default 1, min 0, max 25)
        max_values Optional[int]: The maximum number of items that can be chosen. (default 1, max 25)
        disabled bool: Disable the select and make it not intractable, default false.
        type Union[ComponentType, int]: The action role type number defined by discord. This cannot be modified.
    """

    def __init__(
        self,
        *,
        placeholder: str | None = None,
        min_values: int = 1,
        max_values: int = 1,
        custom_id: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            custom_id=custom_id,
            disabled=disabled,
        )

        self.type: ComponentType = ComponentType.ROLE_SELECT


class MentionableSelectMenu(BaseSelectMenu):
    def __init__(
        self,
        *,
        placeholder: str | None = None,
        min_values: int = 1,
        max_values: int = 1,
        custom_id: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            custom_id=custom_id,
            disabled=disabled,
        )

        self.type: ComponentType = ComponentType.MENTIONABLE_SELECT


class ChannelSelectMenu(BaseSelectMenu):
    def __init__(
        self,
        *,
        channel_types: list[ChannelType] | None = None,
        placeholder: str | None = None,
        min_values: int = 1,
        max_values: int = 1,
        custom_id: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            custom_id=custom_id,
            disabled=disabled,
        )

        self.channel_types: list[ChannelType] | None = channel_types or []
        self.type: ComponentType = ComponentType.CHANNEL_SELECT

    ChannelTypes: ChannelType = ChannelType

    @classmethod
    def from_dict(cls, data: discord_typings.SelectMenuComponentData) -> "ChannelSelectMenu":
        return cls(
            placeholder=data.get("placeholder"),
            min_values=data["min_values"],
            max_values=data["max_values"],
            custom_id=data["custom_id"],
            disabled=data.get("disabled", False),
            channel_types=data.get("channel_types", []),
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type} custom_id={self.custom_id} placeholder={self.placeholder} min_values={self.min_values} max_values={self.max_values} disabled={self.disabled} channel_types={self.channel_types}>"

    def to_dict(self) -> discord_typings.SelectMenuComponentData:
        return {
            **super().to_dict(),
            "channel_types": self.channel_types,
        }


def process_components(
    components: Optional[
        Union[
            List[List[Union[BaseComponent, Dict]]],
            List[Union[BaseComponent, Dict]],
            BaseComponent,
            Dict,
        ]
    ]
) -> List[Dict]:
    """
    Process the passed components into a format discord will understand.

    Args:
        components: List of dict / components to process

    Returns:
        formatted dictionary for discord

    Raises:
        ValueError: Invalid components

    """
    if not components:
        # Its just empty, so nothing to process.
        return components

    if isinstance(components, dict):
        # If a naked dictionary is passed, assume the user knows what they're doing and send it blindly
        # after wrapping it in a list for discord
        return [components]

    if issubclass(type(components), BaseComponent):
        # Naked component was passed
        components = [components]

    if isinstance(components, list):
        if all(isinstance(c, dict) for c in components):
            # user has passed a list of dicts, this is the correct format, blindly send it
            return components

        if all(isinstance(c, list) for c in components):
            # list of lists... actionRow-less sending
            return [ActionRow(*row).to_dict() for row in components]

        if all(issubclass(type(c), InteractiveComponent) for c in components):
            # list of naked components
            return [ActionRow(*components).to_dict()]

        if all(isinstance(c, ActionRow) for c in components):
            # we have a list of action rows
            return [action_row.to_dict() for action_row in components]

    raise ValueError(f"Invalid components: {components}")


def spread_to_rows(*components: Union[ActionRow, Button, StringSelectMenu], max_in_row: int = 5) -> List[ActionRow]:
    """
    A helper function that spreads your components into `ActionRow`s of a set size.

    Args:
        *components: The components to spread, use `None` to explicit start a new row
        max_in_row: The maximum number of components in each row

    Returns:
        List[ActionRow] of components spread to rows

    Raises:
        ValueError: Too many or few components or rows

    """
    # todo: incorrect format errors
    if not components or len(components) > 25:
        raise ValueError("Number of components should be between 1 and 25.")
    return ActionRow.split_components(*components, count_per_row=max_in_row)


def get_components_ids(component: Union[str, dict, list, InteractiveComponent]) -> Iterator[str]:
    """
    Creates a generator with the `custom_id` of a component or list of components.

    Args:
        component: Objects to get `custom_id`s from

    Returns:
        Generator with the `custom_id` of a component or list of components.

    Raises:
        ValueError: Unknown component type

    """
    if isinstance(component, str):
        yield component
    elif isinstance(component, dict):
        if component["type"] == ComponentType.actionrow:
            yield from (comp["custom_id"] for comp in component["components"] if "custom_id" in comp)
        elif "custom_id" in component:
            yield component["custom_id"]
    elif c_id := getattr(component, "custom_id", None):
        yield c_id
    elif isinstance(component, ActionRow):
        yield from (comp_id for comp in component.components for comp_id in get_components_ids(comp))

    elif isinstance(component, list):
        yield from (comp_id for comp in component for comp_id in get_components_ids(comp))
    else:
        raise ValueError(f"Unknown component type of {component} ({type(component)}). " f"Expected str, dict or list")


TYPE_COMPONENT_MAPPING = {
    ComponentType.ACTION_ROW: ActionRow,
    ComponentType.BUTTON: Button,
    ComponentType.STRING_SELECT: StringSelectMenu,
    ComponentType.USER_SELECT: UserSelectMenu,
    ComponentType.CHANNEL_SELECT: ChannelSelectMenu,
    ComponentType.ROLE_SELECT: RoleSelectMenu,
    ComponentType.MENTIONABLE_SELECT: MentionableSelectMenu,
}



from .command import prefixed_command, PrefixedCommand, PrefixedCommandParameter
from .context import PrefixedContext

from .help import PrefixedHelpCommand
from .manager import PrefixedInjectedClient, PrefixedManager, setup
from .utils import when_mentioned, when_mentioned_or

__all__ = (
    "prefixed_command",
    "PrefixedCommand",
    "PrefixedCommandParameter",
    "PrefixedContext",
    "PrefixedHelpCommand",
    "PrefixedInjectedClient",
    "PrefixedManager",
    "setup",
    "when_mentioned",
    "when_mentioned_or",
)



import asyncio
import logging
import os
from asyncio import AbstractEventLoop
from contextlib import suppress
from datetime import datetime

import pytest
import pytest_asyncio

import interactions
from interactions import (
    GuildChannel,
    MessageableMixin,
    GuildNews,
    ThreadableMixin,
    ExplicitContentFilterLevel,
    VerificationLevel,
    Client,
    Guild,
    GuildText,
    Embed,
    PartialEmoji,
    BrandColors,
    Permissions,
    Status,
    EmbedField,
    EmbedAuthor,
    EmbedAttachment,
    EmbedFooter,
    StringSelectOption,
    Modal,
    ParagraphText,
    Message,
    GuildVoice,
)
from interactions.models.discord.asset import Asset
from interactions.models.discord.components import ActionRow, Button, StringSelectMenu
from interactions.models.discord.emoji import process_emoji_req_format
from interactions.api.gateway.websocket import WebsocketClient
from interactions.api.http.route import Route
from interactions.api.voice.audio import AudioVolume
from interactions.client.errors import NotFound
from interactions.models.discord.enums import ButtonStyle
from interactions.models.discord.role import Role

__all__ = ()


try:
    import dotenv

    dotenv.load_dotenv()
except ImportError:
    pass

TOKEN = os.environ.get("BOT_TOKEN")
if not TOKEN:
    pytest.skip(f"Skipping {os.path.basename(__file__)} - no token provided", allow_module_level=True)
if os.environ.get("GITHUB_ACTIONS") and not os.environ.get("RUN_TESTBOT"):
    pytest.skip(f"Skipping {os.path.basename(__file__)} - RUN_TESTBOT not set", allow_module_level=True)

log = logging.getLogger("Interations-Integration-Tests")


@pytest.fixture(scope="module")
def event_loop() -> AbstractEventLoop:
    return asyncio.get_event_loop()


@pytest_asyncio.fixture(scope="module")
async def bot(github_commit) -> Client:
    bot = interactions.Client(activity="Testing someones code")
    await bot.login(TOKEN)
    gw = asyncio.create_task(bot.start_gateway())

    await bot._ready.wait()
    bot.suffix = github_commit
    log.info(f"Logged in as {bot.user} ({bot.user.id}) -- {bot.suffix}")

    yield bot
    gw.cancel()
    with suppress(asyncio.CancelledError):
        await gw


@pytest_asyncio.fixture(scope="module")
async def guild(bot: Client) -> Guild:
    guild = next((g for g in bot.guilds if g.name in ["Interactions.py Test Suite", "NAFF Test Suite"]), None)
    if not guild:
        log.info("No guild found, creating one...")

        guild: interactions.Guild = await interactions.Guild.create("Interactions.py Test Suite", bot)
        community_channel = await guild.create_text_channel("community_channel")

        await guild.edit(
            features=["COMMUNITY"],
            rules_channel=community_channel,
            system_channel=community_channel,
            public_updates_channel=community_channel,
            explicit_content_filter=ExplicitContentFilterLevel.ALL_MEMBERS,
            verification_level=VerificationLevel.LOW,
        )

    yield guild


@pytest_asyncio.fixture(scope="module")
async def channel(bot, guild) -> GuildText:
    channel = await guild.create_text_channel(f"test_scene - {bot.suffix}")
    yield channel
    await channel.delete()


@pytest_asyncio.fixture(scope="module")
async def github_commit() -> str:
    import subprocess

    return subprocess.check_output(["git", "rev-parse", "--short", "HEAD"]).decode("ascii").strip()


def ensure_attributes(target_object) -> None:
    for attr in dir(target_object):
        # ensure all props and attributes load correctly
        getattr(target_object, attr)


@pytest.mark.asyncio
async def test_channels(bot: Client, guild: Guild) -> None:
    channels = [
        guild_category := await guild.create_category("_test_category"),
        await guild.create_text_channel(f"_test_text-{bot.suffix}"),
        await guild.create_news_channel(f"_test_news-{bot.suffix}"),
        await guild.create_stage_channel(f"_test_stage-{bot.suffix}"),
        await guild.create_voice_channel(f"_test_voice-{bot.suffix}"),
    ]

    assert all(c in guild.channels for c in channels)

    channels.append(await bot.owner.fetch_dm())

    try:
        for channel in channels:
            ensure_attributes(channel)

            if isinstance(channel, GuildChannel) and channel != guild_category:
                await channel.edit(parent_id=guild_category.id)
                assert channel.category == guild_category

            if isinstance(channel, MessageableMixin) and not isinstance(channel, GuildVoice):
                _m = await channel.send("test")
                assert _m.channel == channel

                if isinstance(channel, GuildNews):
                    await _m.publish()

                await _m.delete()

            if isinstance(channel, ThreadableMixin):
                if isinstance(channel, GuildNews):
                    _tm = await channel.send("dummy message")
                    thread = await _tm.create_thread("new thread")
                else:
                    thread = await channel.create_thread("new thread")
                assert thread.parent_channel == channel
                _m = await thread.send("test")
                assert _m.channel == thread

                _m = await channel.send("start thread here")
                m_thread = await channel.create_thread("new message thread", message=_m)
                assert _m.id == m_thread.id

                assert m_thread in guild.threads
                assert thread in guild.threads
                await thread.delete()
                # We suppress bcu sometimes event fires too fast, before wait_for is called
                with suppress(asyncio.exceptions.TimeoutError):
                    await bot.wait_for("thread_delete", timeout=2)
                assert thread not in guild.threads
    finally:
        for channel in channels:
            with suppress(NotFound):
                await channel.delete()


@pytest.mark.asyncio
async def test_messages(bot: Client, guild: Guild, channel: GuildText) -> None:
    msg = await channel.send("Message Tests")
    thread = await msg.create_thread("Test Thread")

    try:
        _m = await thread.send("Test")
        ensure_attributes(_m)

        await _m.edit(content="Test Edit")
        assert _m.content == "Test Edit"
        await _m.add_reaction("❌")
        with suppress(asyncio.exceptions.TimeoutError):
            await bot.wait_for("message_reaction_add", timeout=2)

        assert len(_m.reactions) == 1

        assert len(await _m.fetch_reaction("❌")) != 0
        await _m.remove_reaction("❌")
        with suppress(asyncio.exceptions.TimeoutError):
            await bot.wait_for("message_reaction_remove", timeout=2)

        await _m.add_reaction("❌")
        await _m.clear_all_reactions()
        with suppress(asyncio.exceptions.TimeoutError):
            await bot.wait_for("message_reaction_remove_all", timeout=2)

        assert len(_m.reactions) == 0

        await _m.pin()
        assert _m.pinned is True
        await _m.suppress_embeds()
        await _m.unpin()

        _r = await _m.reply(f"test-reply {bot.user.mention} {channel.mention}")
        assert _r._referenced_message_id == _m.id

        mem_mentions = []
        async for member in _r.mention_users:
            mem_mentions.append(member)
        assert len(mem_mentions) == 1

        assert len(_r.mention_channels) == 1

        await thread.send(file=r"tests/LordOfPolls.png")

        assert _m.jump_url is not None
        assert _m.proto_url is not None

        await thread.send(embeds=Embed("Test"))

        await thread.delete()

        _m = await bot.owner.send("Test Message from TestSuite")
        await _m.delete()

    finally:
        with suppress(interactions.errors.NotFound):
            await thread.delete()


@pytest.mark.asyncio
async def test_roles(bot: Client, guild: Guild) -> None:
    roles: list[Role] = []

    try:
        with suppress(interactions.errors.Forbidden):
            roles.append(await guild.create_role("_test_role3"))
            # will fail if the test server has not enabled the feature
            roles.append(await guild.create_role("_test_role1", icon="💥"))
            roles.append(await guild.create_role("_test_role2", icon=r"tests/LordOfPolls.png"))

        assert roles[0].icon is None
        with suppress(IndexError):
            assert isinstance(roles[1].icon, PartialEmoji)
            assert isinstance(roles[2].icon, Asset)
        await guild.me.remove_role(roles[0])

        await roles[0].edit(name="_test_renamed", color=BrandColors.RED)

        assert roles[0].name == "_test_renamed"

        for role in roles:
            await role.delete()

    finally:
        for role in guild.roles:
            if role.name.startswith("_test"):
                with suppress(NotFound):
                    await role.delete()


@pytest.mark.asyncio
async def test_members(bot: Client, guild: Guild, channel: GuildText) -> None:
    member = guild.me
    ensure_attributes(member)

    await member.edit_nickname("Test Nickname")
    with suppress(asyncio.exceptions.TimeoutError):
        await bot.wait_for("member_update", timeout=2)
    assert member.display_name == "Test Nickname"

    await member.edit_nickname(None)
    with suppress(asyncio.exceptions.TimeoutError):
        await bot.wait_for("member_update", timeout=2)
    assert member.display_name == (bot.get_user(member.id)).username

    base_line = len(member.roles)

    assert len(member.roles) == base_line
    role = await guild.create_role(f"test-{bot.suffix}")
    try:
        await member.add_role(role)
        with suppress(asyncio.exceptions.TimeoutError):
            await bot.wait_for("member_update", timeout=2)
        assert len(member.roles) != base_line
        await member.remove_role(role)
        with suppress(asyncio.exceptions.TimeoutError):
            await bot.wait_for("member_update", timeout=2)
        assert len(member.roles) == base_line

        assert member.display_avatar is not None
        assert member.display_name is not None

        assert member.has_permission(Permissions.SEND_MESSAGES)
        assert member.channel_permissions(channel)

        assert member.guild_permissions is not None
    finally:
        await role.delete()


@pytest.mark.asyncio
async def test_gateway(bot: Client) -> None:
    try:
        gateway: WebsocketClient = bot._connection_state.gateway

        assert gateway._entered
        assert gateway._keep_alive is not None

        await bot.change_presence(Status.DO_NOT_DISTURB, activity="Testing")
        await bot.change_presence()

        await gateway.send_heartbeat()
        await gateway._acknowledged.wait()
    finally:
        await bot.change_presence(activity="Testing someones code")


@pytest.mark.asyncio
async def test_ratelimit(bot: Client, channel: GuildText) -> None:
    msg = await channel.send("- Abusing the api... please wait")
    await msg.add_reaction("🤔")
    await msg.remove_reaction("🤔")

    limit = bot.http.get_ratelimit(
        Route(
            "DELETE",
            "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me",
            channel_id=msg.channel.id,
            message_id=msg.id,
            emoji=process_emoji_req_format("🤔"),
        )
    )
    await msg.add_reaction("🤔")
    await msg.remove_reaction("🤔")
    assert limit.locked
    await msg.add_reaction("🤔")
    await msg.remove_reaction("🤔")
    assert limit.locked


@pytest.mark.asyncio
async def test_embeds(bot: Client, channel: GuildText) -> None:
    msg = await channel.send("Embed Tests")
    thread = await msg.create_thread("Test Thread")

    try:
        e = Embed("Test")
        await thread.send(embeds=e)

        e = Embed(
            "Test",
            "Test",
            BrandColors.RED,
            "https://github.com/",
            datetime.now(),
            [
                EmbedField("name", "value"),
                EmbedField("name2", "value2"),
                EmbedField("name3", "value3"),
            ],
            EmbedAuthor(bot.user.display_name, bot.user.avatar.url),
            EmbedAttachment(bot.user.avatar.url),
            EmbedAttachment(bot.owner.avatar.url),
            footer=EmbedFooter("Test", icon_url=bot.user.avatar.url),
        )
        await thread.send(embeds=e)

        e = Embed("Test")
        e.color = BrandColors.RED
        e.url = "https://github.com/"
        e.timestamp = datetime.now()
        e.set_image(bot.user.avatar.url)
        e.set_thumbnail(bot.user.avatar.url)
        e.set_author("Test", bot.owner.avatar.url)
        e.set_footer("Test")
        e.add_field("test", "test")
        e.add_field("test2", "test2")
        await thread.send(embeds=e)

        await thread.delete()
    finally:
        with suppress(interactions.errors.NotFound):
            await thread.delete()


@pytest.mark.asyncio
async def test_components(bot: Client, channel: GuildText) -> None:
    msg = await channel.send("Component Tests")
    thread = await msg.create_thread("Test Thread")

    try:
        await thread.send("Test - single", components=Button(style=ButtonStyle.PRIMARY, label="test"))
        await thread.send(
            "Test - list",
            components=[
                Button(style=ButtonStyle.PRIMARY, label="test"),
                Button(style=ButtonStyle.PRIMARY, label="test"),
            ],
        )
        await thread.send(
            "Test - ActionRow",
            components=ActionRow(
                *[Button(style=ButtonStyle.PRIMARY, label="test"), Button(style=ButtonStyle.PRIMARY, label="test")]
            ),
        )
        await thread.send(
            "Test - StringSelectMenu",
            components=StringSelectMenu(StringSelectOption(label="test", value="test")),
        )

        Modal(ParagraphText(label="test", value="test value, press send"), title="Test Modal")

    finally:
        with suppress(interactions.errors.NotFound):
            await thread.delete()


@pytest.mark.asyncio
async def test_webhooks(bot: Client, guild: Guild, channel: GuildText) -> None:
    test_thread = await channel.create_thread("Test Thread")

    hook = await channel.create_webhook("Test")
    await hook.send("Test 123")
    await hook.delete()

    hook = await channel.create_webhook("Test-Avatar", r"tests/LordOfPolls.png")

    _m = await hook.send("Test", wait=True)
    assert isinstance(_m, Message)
    assert _m.webhook_id == hook.id
    await hook.send("Test", username="Different Name", wait=True)
    await hook.send("Test", avatar_url=bot.user.avatar.url, wait=True)
    _m = await hook.send("Test", thread=test_thread, wait=True)
    assert _m is not None
    assert _m.channel == test_thread

    await hook.delete()


@pytest.mark.asyncio
async def test_voice(bot: Client, guild: Guild) -> None:
    test_channel = await guild.create_voice_channel(f"_test_voice-{bot.suffix}")
    test_channel_two = await guild.create_voice_channel(f"_test_voice_two-{bot.suffix}")
    try:
        try:
            import nacl  # noqa
        except ImportError:
            # testing on a non-voice extra
            return

        vc = await test_channel.connect(deafened=True)
        assert vc == bot.get_bot_voice_state(guild.id)

        audio = AudioVolume("tests/test_audio.mp3")
        vc.play_no_wait(audio)
        await asyncio.sleep(5)

        assert len(vc.current_audio.buffer) != 0
        assert vc.player._sent_payloads != 0

        await vc.move(test_channel_two)
        await asyncio.sleep(2)

        _before = vc.player._sent_payloads

        await test_channel_two.connect(deafened=True)

        await asyncio.sleep(2)

        assert vc.player._sent_payloads != _before

        vc.volume = 1
        await asyncio.sleep(1)
        vc.volume = 0.5

        vc.pause()
        await asyncio.sleep(0.1)
        assert vc.player.paused
        vc.resume()
        await asyncio.sleep(0.1)
        assert not vc.player.paused

        await vc.disconnect()
        await vc._close_connection()
        await vc.ws._closed.wait()
    finally:
        await test_channel.delete()
        await test_channel_two.delete()


@pytest.mark.asyncio
async def test_emoji(bot: Client, guild: Guild) -> None:
    emoji = None
    try:
        emoji = await guild.create_custom_emoji("testEmoji", r"tests/LordOfPolls.png")
        assert emoji.animated is False

        fetched_emoji = await bot.fetch_custom_emoji(emoji.id, guild.id)

        assert emoji == fetched_emoji
        assert emoji.animated == fetched_emoji.animated
        ensure_attributes(emoji)
        ensure_attributes(fetched_emoji)

        await emoji.edit(name="testEditedName")
        await asyncio.sleep(1)
        fetched_emoji = await bot.fetch_custom_emoji(emoji.id, guild.id)
        assert fetched_emoji.name == "testEditedName"

    finally:
        if emoji:
            await emoji.delete()



import copy
from typing import TYPE_CHECKING

import interactions.api.events as events
from interactions.client.const import MISSING
from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("MemberEvents",)


class MemberEvents(EventMixinTemplate):
    @Processor.define()
    async def _on_raw_guild_member_add(self, event: "RawGatewayEvent") -> None:
        g_id = event.data.pop("guild_id")
        member = self.cache.place_member_data(g_id, event.data)
        if guild := self.cache.get_guild(g_id):
            guild.member_count += 1
        self.dispatch(events.MemberAdd(g_id, member))

    @Processor.define()
    async def _on_raw_guild_member_remove(self, event: "RawGatewayEvent") -> None:
        g_id = event.data.pop("guild_id")
        user = self.cache.place_user_data(event.data["user"])
        member = self.cache.get_member(g_id, user.id)

        self.cache.delete_member(g_id, user.id)
        if guild := self.cache.get_guild(g_id):
            guild.member_count -= 1

        self.dispatch(events.MemberRemove(g_id, member or user))

    @Processor.define()
    async def _on_raw_guild_member_update(self, event: "RawGatewayEvent") -> None:
        g_id = event.data.pop("guild_id")
        before = copy.copy(self.cache.get_member(g_id, event.data["user"]["id"])) or MISSING
        self.dispatch(events.MemberUpdate(g_id, before, self.cache.place_member_data(g_id, event.data)))
